from bidict import bidict

nsmap = bidict(
    {
        "gen": "http://www.opengis.net/citygml/generics/2.0",
        "gml": "http://www.opengis.net/gml",
        "core": "http://www.opengis.net/citygml/2.0",
        "bldg": "http://www.opengis.net/citygml/building/2.0",
        "tran": "http://www.opengis.net/citygml/transportation/2.0",
        "brid": "http://www.opengis.net/citygml/bridge/2.0",
        "xlink": "http://www.w3.org/1999/xlink",
        "xAL": "urn:oasis:names:tc:ciq:xsdschema:xAL:2.0",
        "uro": "https://www.geospatial.jp/iur/uro/2.0",
        "dem": "http://www.opengis.net/citygml/relief/2.0",
        "frn": "http://www.opengis.net/citygml/cityfurniture/2.0",
        "urf": "https://www.geospatial.jp/iur/urf/2.0",
        "luse": "http://www.opengis.net/citygml/landuse/2.0",
        "veg": "http://www.opengis.net/citygml/vegetation/2.0",
    }
)

prefectures = [
    "北海道",
    "青森県",
    "岩手県",
    "宮城県",
    "秋田県",
    "山形県",
    "福島県",
    "茨城県",
    "栃木県",
    "群馬県",
    "埼玉県",
    "千葉県",
    "東京都",
    "神奈川県",
    "新潟県",
    "富山県",
    "石川県",
    "福井県",
    "山梨県",
    "長野県",
    "岐阜県",
    "静岡県",
    "愛知県",
    "三重県",
    "滋賀県",
    "京都府",
    "大阪府",
    "兵庫県",
    "奈良県",
    "和歌山県",
    "鳥取県",
    "島根県",
    "岡山県",
    "広島県",
    "山口県",
    "徳島県",
    "香川県",
    "愛媛県",
    "高知県",
    "福岡県",
    "佐賀県",
    "長崎県",
    "熊本県",
    "大分県",
    "宮崎県",
    "鹿児島県",
    "沖縄県",
]



import re
import sys
import zipfile
from os import PathLike
from pathlib import Path, PurePosixPath
from typing import Literal, Sequence

import flatgeobuf as fgb
import geopandas as gpd

try:
    from pyogrio import read_dataframe
except ImportError:
    read_dataframe = None

from plateaukit import defaults, geocoding
from plateaukit.config import Config
from plateaukit.core.area import Area
from plateaukit.logger import logger


class Dataset:
    """This class represents a dataset.

    Attributes:
        dataset_id: Dataset ID.
        gdf: The GeoDataFrame of the dataset.
    """

    dataset_id: str
    gdf: gpd.GeoDataFrame | None

    def __init__(self, dataset_id: str):
        """Initialize a dataset.

        Args:
            dataset_id: Dataset ID
        """

        self.dataset_id = dataset_id
        self.gdf = None

        # config = Config()
        # gpkg_path = config.data[dataset_id].get("gpkg")
        # self.gdf = read_dataframe(gpkg_path)

    def __repr__(self):
        return f"Dataset({self.dataset_id})"

    def load_gdf(self, format: Literal["parquet", "gpkg"] = "parquet"):
        """Load the GeoDataFrame from the prebuilt dataset."""

        # NOTE: Fallback to GeoPackage
        def load_gdf_gpkg():
            gpkg_path = config.datasets[self.dataset_id].get("gpkg")

            if gpkg_path:
                if not read_dataframe:
                    raise ImportError(
                        "Package pyogrio is required. Please install it using `pip install pyogrio`."
                    ) from None
                self.gdf = read_dataframe(gpkg_path)
            else:
                raise RuntimeError(
                    "Missing GeoPackage; Please prebuild the dataset first"
                )

        config = Config()

        if format == "parquet":
            logger.debug("Loading Parquet file...")
            parquet_path = config.datasets[self.dataset_id].get("parquet")

            if parquet_path:
                self.gdf = gpd.read_parquet(parquet_path)
            else:
                load_gdf_gpkg()

        elif format == "gpkg":
            load_gdf_gpkg()

        else:
            raise ValueError("Invalid format")

    def get_area(self):
        """Get the entire area of the dataset."""

        if self.dataset_id.endswith(".cloud"):
            raise Exception("Not supported for cloud datasets")

        if self.gdf is None:
            self.load_gdf()

        area_gdf = self.gdf

        area = Area(area_gdf)
        area._datasets = [self.dataset_id]

        return area

    def area_from_bbox(self, bbox: list[float] | None = None):
        """Get the specified area from the dataset.

        Args:
            bbox: Bounding box of the area of interest.
                   If not specified, the area of the entire dataset will be returned.
        """

        if self.dataset_id.endswith(".cloud"):
            if "pyodide" in sys.modules:
                try:
                    import pyodide_http  # type: ignore

                    pyodide_http.patch_all()
                except Exception:
                    pass

            # Load data from remote flatgeobuf
            remote_fgb = defaults.cloud_base_url + self.dataset_id.replace(
                ".cloud", ".fgb"
            )
            if read_dataframe:
                area_gdf = read_dataframe(
                    remote_fgb,
                    bbox=tuple(bbox),
                )
            else:
                feature_collection = fgb.load_http(
                    remote_fgb,
                    bbox=tuple(bbox),
                )
                area_gdf = gpd.GeoDataFrame.from_features(feature_collection)
        else:
            if self.gdf is None:
                self.load_gdf()

            area_gdf = (
                self.gdf.cx[bbox[0] : bbox[2], bbox[1] : bbox[3]] if bbox else self.gdf
            )

        # TODO: Error handling when area_gdf is empty

        area = Area(area_gdf)
        area._datasets = [self.dataset_id]

        return area

    def area_from_polygons(self, polygons: list[gpd.GeoDataFrame]):
        """Get an area from the dataset by polygons.

        Args:
            polygon: Polygon of the area of interest.
        """

        if self.dataset_id.endswith(".cloud"):
            raise Exception("Not supported for cloud datasets")

        if self.gdf is None:
            self.load_gdf()

        area_gdf = self.gdf[self.gdf.geometry.intersects(polygons)]

        return Area(area_gdf)

    def area_from_points(
        self, points: list[Sequence[float]], size: list[float] = [1000, 1000]
    ):
        """Get an area from the dataset by points.

        Args:
            points: List of points of the area of interest.
            size: Size of the area of interest in meters.
        """

        bbox = geocoding._get_bbox(points)
        bbox = geocoding._pad_bbox(bbox, size)

        return self.area_from_bbox(bbox)

    def area_from_address(self, address: str, size: list[float] = [1000, 1000]):
        """Get an area from the dataset by an address.

        Args:
            address: Address of the area of interest.
            size: Size of the area of interest in meters.
        """

        bbox = geocoding.bbox_from_address(address, min_size=size)

        return self.area_from_bbox(bbox)

    def area_from_postcode(self, postcode: str):
        """Get an area from the dataset by a postcode.

        Args:
            postcode: Postal code of the area of interest.
        """

        bbox = geocoding.bbox_from_postcode(postcode)

        return self.area_from_bbox(bbox)

    def area_from_landmark(self, landmark: str, min_size: list[float] = [1000, 1000]):
        """Get an area from the dataset by a landmark.

        Args:
            landmark: Landmark of the area of interest.
        """
        bbox = geocoding.bbox_from_landmark(landmark, min_size=min_size)

        return self.area_from_bbox(bbox)

    # def search_area(self, keyword: str):
    #     print(keyword)

    def to_geojson(
        self,
        outfile: str | PathLike,
        *,
        types: list[str] = ["bldg"],
        altitude: bool = True,
        include_type: bool = False,
        seq=False,
        split: int = 1,
        **kwargs,
    ):
        """Generate GeoJSON from PLATEAU datasets.

        Args:
            outfile: Output file path.
            types: CityGML object types to include.
            split: Split the output into specified number of files.
            **kwargs: Keyword arguments for the generator.
        """
        # NOTE: generators requires multiprocessing at the moment, unavailable in pyodide
        from plateaukit import generators

        if not self.dataset_id:
            raise Exception("Missing dataset_id")

        if not types:
            raise Exception("Missing object types")

        # NOTE: this is intentional but to be refactored in the future
        config = Config()
        record = config.datasets[self.dataset_id]

        if "citygml" not in record:
            raise Exception("Missing CityGML data")
        file_path = Path(record["citygml"])

        infiles = []

        # TODO: Refactor
        for type in types:
            # TODO: Fix
            pat = re.compile(rf".*udx\/{type}\/.*\.gml$")
            if zipfile.is_zipfile(file_path):
                with zipfile.ZipFile(file_path) as f:
                    namelist = f.namelist()
                    targets = list(filter(lambda x: pat.match(x), namelist))
                    # print(targets)

                    # NOTE: zipfs requires POSIX path
                    infiles += [str(PurePosixPath("/", target)) for target in targets]

                    # f.extractall(tdir, members=targets)
                    # infiles += [
                    #     str(Path(tdir, Path(file_path).stem, "udx", type, "*.gml"))
                    # ]
            else:
                infiles += [str(Path(file_path, "udx", type, "*.gml"))]

        logger.debug([types, infiles, outfile])

        # Codelists
        codelist_infiles = []
        pat = re.compile(r".*codelists\/.*\.xml$")

        if zipfile.is_zipfile(file_path):
            with zipfile.ZipFile(file_path) as f:
                namelist = f.namelist()
                codelist_infiles = list(filter(lambda x: pat.match(x), namelist))
                # NOTE: zipfs requires POSIX path
                codelist_infiles = [
                    str(PurePosixPath("/", target)) for target in codelist_infiles
                ]
        else:
            # TODO: Test support for non-zip codelists
            codelist_infiles += [str(Path(file_path, "codelists", "*.xml"))]

        generators.geojson.geojson_from_citygml(
            infiles,
            outfile,
            types=types,
            altitude=altitude,
            include_type=include_type,
            split=split,
            seq=seq,
            zipfile=file_path,
            codelist_infiles=codelist_infiles,
            **kwargs,
        )

    def to_cityjson(
        self,
        outfile: str | PathLike,
        *,
        types: list[str] = ["bldg"],
        object_types=None,  # TODO: Handle this
        lod=[1, 2],
        ground: bool = False,
        seq: bool = False,
        split: int = 1,
        selection: list[str] | None = None,
        target_epsg: int | None = None,
        **kwargs,
    ):
        """Generate CityJSON from PLATEAU datasets.

        Args:
            outfile: Output file path.
            types: CityGML feature types.
            split: Split the output into specified number of files.
            seq: Generate CityJSONSeq.
            **kwargs: Keyword arguments for the generator.
        """
        # NOTE: generators requires multiprocessing at the moment, unavailable in pyodide
        from plateaukit import generators

        params = {}

        # if precision:
        #     params["precision"] = precision

        if not self.dataset_id:
            raise Exception("Missing dataset_id")

        config = Config()
        record = config.datasets[self.dataset_id]

        if "citygml" not in record:
            raise Exception("Missing CityGML data")

        file_path = Path(record["citygml"])

        infiles = []

        for type in types:
            # TODO: fix
            pat = re.compile(rf".*udx\/{type}\/.*\.gml$")

            if zipfile.is_zipfile(file_path):
                with zipfile.ZipFile(file_path) as f:
                    namelist = f.namelist()
                    targets = list(filter(lambda x: pat.match(x), namelist))
                    # NOTE: zipfs requires POSIX path
                    infiles += [str(PurePosixPath("/", target)) for target in targets]
            else:
                infiles += [str(Path(file_path, "udx", type, "*.gml"))]

            logger.debug([types, infiles, outfile])

        # Sort by a part of filename to group by areas; TODO: Update this
        # try:
        #     infiles = sorted(infiles, key=lambda x: Path(x).stem.split("_")[0])
        # except:
        #     infiles = sorted(infiles)
        infiles = sorted(infiles)

        # Codelists
        codelist_infiles = []
        pat = re.compile(r".*codelists\/.*\.xml$")

        if zipfile.is_zipfile(file_path):
            with zipfile.ZipFile(file_path) as f:
                namelist = f.namelist()
                codelist_infiles = list(filter(lambda x: pat.match(x), namelist))
                # NOTE: zipfs requires POSIX path
                codelist_infiles = [
                    str(Path("/", target)) for target in codelist_infiles
                ]
        else:
            # TODO: Test support for non-zip codelists
            codelist_infiles += [str(Path(file_path, "codelists", "*.xml"))]

        generators.cityjson.cityjson_from_citygml(
            infiles,
            outfile,
            split=split,
            zipfile=file_path,
            lod=lod,
            ground=ground,
            seq=seq,
            codelist_infiles=codelist_infiles,
            selection=selection,
            target_epsg=target_epsg,
            **kwargs,
        )


def load_dataset(dataset_id: str) -> Dataset:
    """Load a dataset.

    Args:
        dataset_id: Dataset ID.

    Returns:
        Dataset: Dataset object.
    """
    return Dataset(dataset_id)



import io
import os.path
from pathlib import Path

import geojson
from fs import open_fs
from geojson import FeatureCollection

from plateaukit.logger import logger
from ._single import geojson_from_gml_single, features_from_gml_single


def geojson_from_gml_serial_with_quit(
    infiles,
    outfile,
    *,
    seq=False,
    codelist_infiles=None,
    types=None,
    include_type=False,
    zipfile=None,
    task_id=None,
    _quit=None,
    _progress=None,
    **kwargs,
):
    """Generate GeoJSON from multiple CityGML files."""
    # Load codelists
    codelist_file_map = None

    if zipfile is not None:
        zip_fs = open_fs(f"zip://{zipfile}")
    else:
        zip_fs = None

    if codelist_infiles and infiles:
        base_path = Path(infiles[0]).parent  # TODO: Fix this
        codelist_file_map = dict()

        for codelist_infile in codelist_infiles:
            if zip_fs is not None:
                with zip_fs.openbin(codelist_infile, "r") as f:
                    relative_path = os.path.relpath(codelist_infile, base_path)
                    codelist_file_map[relative_path] = io.BytesIO(f.read())
            else:
                with open(codelist_infile, "rb") as f:
                    relative_path = os.path.relpath(codelist_infile, base_path)
                    codelist_file_map[relative_path] = io.BytesIO(f.read())

    total = len(infiles) + 1  # + 1 for geojson.dump

    if seq:
        # NOTE: Precision https://github.com/jazzband/geojson#default-and-custom-precision
        with open(outfile, "w") as obuf:
            for i, infile in enumerate(infiles):
                if task_id is not None and _progress is not None:
                    _progress[task_id] = {"progress": i + 1, "total": total}

                if _quit and _quit.is_set():
                    return

                logger.debug(f"infile: {infile}")

                _open = zip_fs.openbin if zip_fs else open

                with _open(infile, "rb") as inbuf:
                    infile_features = features_from_gml_single(
                        inbuf,
                        types=types,
                        codelist_file_map=codelist_file_map,
                        include_type=include_type,
                        **kwargs,
                    )
                    for feature in infile_features:
                        geojson.dump(
                            feature, obuf, ensure_ascii=False, separators=(",", ":")
                        )
                        obuf.write("\n")

    else:
        features = []

        for i, infile in enumerate(infiles):
            if task_id is not None and _progress is not None:
                _progress[task_id] = {"progress": i + 1, "total": total}

            if _quit and _quit.is_set():
                return

            logger.debug(f"infile: {infile}")

            _open = zip_fs.openbin if zip_fs else open

            with _open(infile, "rb") as f:
                infile_features = list(
                    features_from_gml_single(
                        f,
                        types=types,
                        codelist_file_map=codelist_file_map,
                        include_type=include_type,
                        **kwargs,
                    )
                )

            try:
                features.extend(features)
            except Exception as err:
                logger.debug(err)

        collection = FeatureCollection(features)

        # NOTE: Precision https://github.com/jazzband/geojson#default-and-custom-precision
        with open(outfile, "w") as f:
            geojson.dump(collection, f, ensure_ascii=False, separators=(",", ":"))

    if zip_fs:
        zip_fs.close()

    # Complete progress
    if _progress:
        _progress[task_id] = {"progress": total, "total": total}



from pathlib import Path

from plateaukit.logger import logger


def geojson_to_gpkg(
    infiles,
    outfile,
):
    import geopandas as gpd

    for infile in infiles:
        logger.debug(infile)
        with open(infile, "r") as f:
            # print(infile)
            df = gpd.read_file(f, encoding="tuf-8")
            # print(df)
            if not Path(outfile).exists():
                df.to_file(outfile, driver="GPKG", mode="w")
            else:
                df.to_file(outfile, driver="GPKG", mode="a")



from plateaukit.formats.citygml.constants import nsmap
from plateaukit.formats.citygml.parsers.xml_models.city_object import CityObjectXML


def _get_string_attribute(xml: CityObjectXML, name) -> str | None:
    path = f"./gen:stringAttribute[@name='{name}']/gen:value"
    result = xml.find(path, nsmap)
    return result.text if result is not None else None


def get_name(xml: CityObjectXML) -> str | None:
    value = xml._get_codespace_attribute("./gml:name")
    return value


def get_usage(xml: CityObjectXML) -> str | None:
    value = xml._get_codespace_attribute("./bldg:usage")
    return value


def get_building_id(xml: CityObjectXML) -> str | None:
    try:
        path = "./uro:buildingIDAttribute/uro:BuildingIDAttribute/uro:buildingID"
        result = xml.find(path, nsmap)
        value = result.text if result is not None else None
        assert value is not None
        return value
    except AssertionError:
        pass

    try:
        value = _get_string_attribute(xml, name="建物ID")
        assert value is not None
        return value
    except AssertionError:
        pass

    return None


def get_measured_height(xml: CityObjectXML) -> float | None:
    result = xml.find("./bldg:measuredHeight", nsmap)
    value = result.text if result is not None else None
    value = float(value) if value is not None else None
    return value


def get_year_of_construction(xml: CityObjectXML) -> int | None:
    result = xml.find("./bldg:yearOfConstruction", nsmap)
    value = result.text if result is not None else None
    value = int(value) if value is not None else None
    return value


def get_storeys_above_ground(xml: CityObjectXML) -> int | None:
    result = xml.find("./bldg:storeysAboveGround", nsmap)
    value = result.text if result is not None else None
    value = int(value) if value is not None else None
    return value


def get_storeys_below_ground(xml: CityObjectXML) -> int | None:
    result = xml.find("./bldg:storeysBelowGround", nsmap)
    value = result.text if result is not None else None
    value = int(value) if value is not None else None
    return value


def get_address(xml: CityObjectXML) -> dict | None:
    el = xml.find("./bldg:address", nsmap)

    if el is None:
        return None

    result = el.find(".//xAL:LocalityName", nsmap)

    if result is None:
        return None

    locality_name = result.text

    addr = {
        "locality_name": locality_name,
    }

    return addr



# %%
from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Annotated, Optional, Sequence

import requests
from normalize_japanese_addresses import normalize
from pyproj import Transformer
from pyproj.enums import TransformDirection

wgs84_to_sm = Transformer.from_crs("epsg:4326", "epsg:3857", always_xy=True)


@dataclass
class PostalAddress:
    prefecture: str
    address1: str
    address2: str
    address3: str
    address4: str


@dataclass
class AddressPoint:
    lat: Optional[float]
    lng: Optional[float]
    prefecture: str
    city: str
    town: str
    addr: Optional[str]


@dataclass
class LandmarkPoint:
    lat: float
    lng: float
    name: str


@dataclass
class ResultPoint:
    lng: float
    lat: float
    all: Optional[list[AddressPoint | LandmarkPoint]] = None


def _get_bbox(
    points: list[Annotated[Sequence[float], 2] | Annotated[Sequence[float], 3]]
):
    """Returns a bounding box from a list of points."""
    xs = [p[0] for p in points]
    ys = [p[1] for p in points]

    return [min(xs), min(ys), max(xs), max(ys)]


def _get_bbox_centroid(bbox: list[float]):
    """Returns a centroid from a bounding box."""
    return [(bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2]


def _pad_bbox(bbox: list[float], min_size: Annotated[Sequence[float], 2]):
    """Returns a padded bounding box."""

    # print(bbox)

    bbox_sm = wgs84_to_sm.itransform([(bbox[0], bbox[1]), (bbox[2], bbox[3])])
    bbox_sm = list(bbox_sm)

    top_left, bottom_right = list(map(list, bbox_sm))

    # print(bbox_sm)

    if top_left[0] - bottom_right[0] < min_size[0]:
        top_left[0] -= (min_size[0] - (top_left[0] - bottom_right[0])) / 2
        bottom_right[0] += (min_size[0] - (top_left[0] - bottom_right[0])) / 2
    if top_left[1] - bottom_right[1] < min_size[1]:
        top_left[1] -= (min_size[1] - (top_left[1] - bottom_right[1])) / 2
        bottom_right[1] += (min_size[1] - (top_left[1] - bottom_right[1])) / 2

    bbox_wgs84 = wgs84_to_sm.itransform(
        [top_left, bottom_right], direction=TransformDirection.INVERSE
    )
    bbox_wgs84 = list(map(list, bbox_wgs84))
    bbox_wgs84 = sum(bbox_wgs84, [])
    # print(bbox_wgs84)

    return bbox_wgs84


def address_from_postcode(code: str):
    """Returns an address from a Japanese postal code.

    >>> address_from_postcode("108-0073")
    PostalAddress(prefecture='東京都', address1='港区', address2='三田', address3='', address4='')
    """

    parsed = re.match(r"^([0-9]{3})\-?([0-9]{4})$", code)

    if not parsed:
        raise RuntimeError("Invalid postal code")

    first, second = parsed.groups()
    # Using a fork by arrow-payment of https://madefor.github.io/postal-code-api/api/v1/ . Thanks!
    resp = requests.get(
        f"https://arrow-payment.github.io/postal-code-api/api/v2/{first}/{second}.json"
    )

    if resp.status_code != 200:
        raise RuntimeError("Address not found")

    body = resp.json()
    data = body["data"]

    # # TODO: Support multiple addresses

    try:
        return PostalAddress(
            data[0]["ja"]["prefecture"],
            data[0]["ja"]["address1"],
            data[0]["ja"]["address2"],
            data[0]["ja"]["address3"],
            data[0]["ja"]["address4"],
        )
    except:
        raise RuntimeError()


def point_from_address(address: str):
    """Returns a point from an address.

    >>> from_address("大阪府大阪市北区梅田２丁目２−２")
    Address(prefecture='大阪府', city='大阪市北区', town='梅田二丁目', addr='2-2', lat=34.698571, lng=135.49385)
    """
    normalized_addr = normalize(address)

    return AddressPoint(
        lat=normalized_addr["lat"],
        lng=normalized_addr["lng"],
        prefecture=normalized_addr["pref"],
        city=normalized_addr["city"],
        town=normalized_addr["town"],
        addr=normalized_addr["addr"],
    )


def bbox_from_address(
    address: str, min_size: Annotated[Sequence[float], 2] = [1000, 1000]
):
    """Returns a bounding box from an address."""

    point = point_from_address(address)
    bbox = _get_bbox([(point.lng, point.lat)])
    bbox = _pad_bbox(bbox, min_size)
    return bbox


def point_from_postcode(code: str):
    """Returns a coordinate from a Japanese postal code.

    >>> from_postcode("108-0073")  # doctest: +ELLIPSIS
    Result(lng=139.74034749999998, lat=35.648897000000005, all=[...])
    """
    postal_address = address_from_postcode(code)
    resp = requests.get(
        f"https://geolonia.github.io/japanese-addresses/api/ja/{postal_address.prefecture}/{postal_address.address1}.json"
    )

    if resp.status_code != 200:
        raise RuntimeError("Address not found")

    city_data = resp.json()

    if postal_address.address2:
        town_data = list(
            filter(lambda x: x["town"].startswith(postal_address.address2), city_data)
        )
        # print(town_data)

        bbox = _get_bbox([(x["lng"], x["lat"]) for x in town_data])
        # print(bbox)

        centroid = _get_bbox_centroid(bbox)
        # print(centroid)

        return ResultPoint(
            lng=centroid[0],
            lat=centroid[1],
            all=[
                AddressPoint(
                    lat=x["lat"],
                    lng=x["lng"],
                    prefecture=postal_address.prefecture,
                    city=postal_address.address1,
                    town=x["town"],
                    addr=None,
                )
                for x in town_data
            ],
        )
    else:
        raise NotImplementedError()


def bbox_from_postcode(
    code: str, min_size: Annotated[Sequence[float], 2] = [1000, 1000]
):
    """Returns a bounding box from a Japanese postal code."""

    point = point_from_postcode(code)
    bbox = _get_bbox([(point.lng, point.lat)])
    bbox = _pad_bbox(bbox, min_size)
    return bbox


def point_from_landmark(landmark_name: str):
    """Returns a coordinate from a landmark.

    >>> from_landmark("東京タワー")
    Result(lng=139.745555555, lat=35.658611111, all=[Landmark(name='東京タワー', lat=139.745555555, lng=35.658611111)])
    """
    # Query Wikidata for the coordinate from a landmark name.
    query = """SELECT distinct ?item ?itemLabel ?coordinate_location WHERE {{
        ?item ?label "{label}"@ja;
              wdt:P625 ?coordinate_location.
        ?article schema:about ?item .
        ?article schema:inLanguage "ja" .
        ?article schema:isPartOf <https://ja.wikipedia.org/>.
        SERVICE wikibase:label {{ bd:serviceParam wikibase:language "ja". }}
    }}
    """.format(
        label=landmark_name
    )

    url = "https://query.wikidata.org/sparql"

    r = requests.get(url, params={"format": "json", "query": query})
    data = r.json()

    # print(data)

    records = data["results"]["bindings"]

    if len(records) == 0:
        return None

    def parse_point(literal):
        m = re.match(r"^Point\(([0-9\.]+) ([0-9\.]+)\)$", literal)
        if not m:
            raise RuntimeError("Failed to parse coordinate")

        lng, lat = m.groups()

        return [float(lng), float(lat)]

    coordinate_raw = records[0]["coordinate_location"]["value"]
    lng, lat = parse_point(coordinate_raw)

    return ResultPoint(
        lng=lng,
        lat=lat,
        all=[
            LandmarkPoint(
                *parse_point(record["coordinate_location"]["value"]),
                record["itemLabel"]["value"],
            )
            for record in records
        ],
    )


def bbox_from_landmark(
    landmark_name: str, min_size: Annotated[Sequence[float], 2] = [1000, 1000]
):
    """Returns a bounding box from a landmark."""

    point = point_from_landmark(landmark_name)

    if point is None:
        return None

    bbox = _get_bbox([(point.lng, point.lat)])
    bbox = _pad_bbox(bbox, min_size)

    return bbox



import pyproj
from lxml import etree

from plateaukit import utils
from plateaukit.formats.citygml.constants import nsmap

MAPLIBRE_SCALE_FACTOR = 10000000
MERCATOR_HALF_WORLD_LENGTH = 20037508.342789243906736373901367187500


Vector3 = tuple[float, float, float]
Point = Vector3


class GeometryParser:
    """A parser for CityGML geometries.

    Attributes:
        transformer: A pyproj.Transformer instance.
        maplibre: Whether to use MapLibre scale factor.
    """

    transformer: pyproj.Transformer | None
    maplibre: bool

    def __init__(
        self, transformer: pyproj.Transformer | None = None, maplibre: bool = False
    ):
        self.transformer = transformer
        self.maplibre = maplibre

    def _transform(
        self,
        vertices: list[Point],
        scale: Vector3 = (1, 1, 1),
        translate: Vector3 = (0, 0, 0),
    ) -> list[Point]:
        # TODO: numpy

        transformed = []

        for vertex in vertices:
            x, y, z = vertex
            nx = (x * scale[0]) + translate[0]
            ny = (y * scale[1]) + translate[1]
            nz = (z * scale[2]) + translate[2]
            transformed.append((nx, ny, nz))

        return transformed

    def _adjust_mercator_to_maplibre(self, vertices):
        xy = -(0.5 * MAPLIBRE_SCALE_FACTOR) / MERCATOR_HALF_WORLD_LENGTH
        z = 0.5 * MAPLIBRE_SCALE_FACTOR / MERCATOR_HALF_WORLD_LENGTH
        scale = (xy, xy, z)

        transformed = self._transform(
            vertices,
            scale=scale,
            translate=(
                0.5 * MAPLIBRE_SCALE_FACTOR,
                0.5 * MAPLIBRE_SCALE_FACTOR,
                0,
            ),
        )

        return transformed

    def extract_chunked_poslists(self, root: etree._Element):
        path = "/".join(
            [
                ".//gml:surfaceMember",  # TODO: Check this
                "gml:Polygon",
                "gml:exterior",
                "gml:LinearRing",
                "gml:posList",
            ]
        )
        # print(path)
        # results = root.findall(path, nsmap)

        parsed = []

        for result in root.iterfind(path, nsmap):
            poslist = map(float, result.text.split(" "))

            chunked = list(utils.chunker(poslist, 3))

            if self.maplibre and self.transformer:
                chunked = list(self.transformer.itransform(chunked))
                chunked = self._adjust_mercator_to_maplibre(chunked)
            elif self.transformer:
                chunked = list(self.transformer.itransform(chunked))

            surface = [chunked]
            parsed.append(surface)

        return parsed



import glob
import tempfile
from pathlib import Path
from typing import Literal

import geopandas as gpd
import pandas as pd
from rich import get_console

from plateaukit.config import Config
from plateaukit.core.dataset import load_dataset
from plateaukit.logger import logger


def prebuild(
    dataset_id: str,
    *,
    split: int = 10,
    simple_output=False,
    format: Literal["gpkg", "parquet"] = "parquet",
) -> None:
    """Prebuild PLATEAU datasets."""

    try:
        from pyogrio import read_dataframe, write_dataframe
    except ImportError:
        raise ImportError(
            "Package pyogrio is required. Please install it using `pip install pyogrio`."
        ) from None

    console = get_console()

    if not dataset_id:
        raise Exception("Missing argument: dataset_id")

    config = Config()
    record = config.datasets.get(dataset_id)
    # print(dataset_id, record)

    # TODO: All types
    types = ["bldg"]

    with tempfile.TemporaryDirectory() as tdir:
        logger.debug(f"Temporary directory: {tdir}")

        for type in types:
            outfile = Path(tdir, f"{dataset_id}.{type}.geojsonl")

            if dataset_id:
                dataset = load_dataset(dataset_id)
                dataset.to_geojson(
                    outfile,
                    types=types,
                    altitude=False,  # TODO: Check this
                    include_type=True,
                    seq=True,
                    split=split,
                    progress={"description": "Generating GeoJSONSeq files..."},
                    simple_output=simple_output,
                )
            else:
                raise NotImplementedError()

        if format == "parquet":
            display_name = "Parquet files"
        elif format == "gpkg":
            display_name = "GeoPackage files"
        else:
            raise NotImplementedError()

        with console.status(f"Writing {display_name}...") as status:
            if format == "gpkg":
                df = gpd.GeoDataFrame()

                for filename in glob.glob(str(Path(tdir, "*.geojsonl"))):
                    subdf = read_dataframe(filename)
                    df = pd.concat([df, subdf])

                # TODO: Use more accurate CRS
                mercator = df.to_crs(3857)
                centroid_mercator = mercator.centroid
                centroid = centroid_mercator.to_crs(4326)

                df["longitude"] = centroid.x
                df["latitude"] = centroid.y

                dest_path = Path(config.data_dir, f"{dataset_id}.gpkg")
                write_dataframe(df, dest_path, driver="GPKG")

                config.datasets[dataset_id]["gpkg"] = dest_path
                config.save()
            elif format == "parquet":
                tmp_file_paths = []

                for filename in glob.glob(str(Path(tdir, "*.geojsonl"))):
                    df = gpd.read_file(filename)

                    # TODO: Use more accurate CRS
                    mercator = df.to_crs(3857)
                    centroid_mercator = mercator.centroid
                    centroid = centroid_mercator.to_crs(4326)

                    df["longitude"] = centroid.x
                    df["latitude"] = centroid.y

                    tmp_dest_path = str(Path(tdir, f"{filename}.parquet"))
                    df.to_parquet(tmp_dest_path)

                    tmp_file_paths.append(tmp_dest_path)

                df = None
                for filename in tmp_file_paths:
                    subdf = gpd.read_parquet(filename)
                    df = pd.concat([df, subdf]) if df is not None else subdf

                if df is None:
                    raise RuntimeError

                dest_path = Path(config.data_dir, f"{dataset_id}.parquet")
                df.to_parquet(dest_path, compression="zstd")

                config.datasets[dataset_id]["parquet"] = dest_path
                config.save()

            else:
                raise NotImplementedError()

        console.print(f"Writing {display_name}... [green]Done")

        # click.echo(f"\nCreated: {dest_path}")



import tempfile

from geopandas import GeoDataFrame


class Building:
    """This class represents a building."""

    gdf: GeoDataFrame

    def __init__(self, gdf: GeoDataFrame) -> None:
        """Initialize a building.

        Args:
            gdf: GeoDataFrame representing the building.
        """
        self.gdf = gdf

    def show(self):
        import pydeck

        # bbox = self.gdf.total_bounds
        bbox = self.gdf.buffer(0.001).total_bounds
        points = [(bbox[0], bbox[1]), (bbox[2], bbox[3])]
        # print(points)

        view_state = pydeck.data_utils.compute_view(points, view_proportion=1)
        view_state.pitch = 45

        deck = pydeck.Deck(
            layers=[
                pydeck.Layer(
                    "GeoJsonLayer",
                    data=self.gdf,
                    filled=True,
                    get_fill_color=[255, 255, 255],
                    extruded=True,
                    get_elevation="measuredHeight",
                    pickable=True,
                )
            ],
            initial_view_state=view_state,
            tooltip={
                "html": "<b>ID:</b> {id}",
                "style": {
                    "font-family": "sans-serif",
                    "font-size": "8px",
                    "color": "white",
                },
            },
        )

        try:
            if __IPYTHON__:  # type: ignore
                return deck.to_html()
        except:
            pass

        with tempfile.NamedTemporaryFile(suffix=".html", delete=False) as fp:
            deck.to_html(fp.name, open_browser=True)

    def geojson(self):
        raise NotImplementedError()

    def cityjson(self):
        raise NotImplementedError()

    @property
    def attributes(self):
        """Return the attributes of the building."""

        # TODO:
        # - lng, lat

        return {
            "id": self.gdf.id.iloc[0],
            "measuredHeight": self.gdf.measuredHeight.iloc[0],
            # "roofType": self.gdf.roofType.iloc[0],
            # "roofShape": self.gdf.roofShape.iloc[0],
            # "roofMaterial": self.gdf.roofMaterial.iloc[0],
            # "roofColor": self.gdf.roofColor.iloc[0],
            # "facadeMaterial": self.gdf.facadeMaterial.iloc[0],
            # "facadeColor": self.gdf.facadeColor.iloc[0],
            # "buildingType": self.gdf.buildingType.iloc[0],
        }

    def __repr__(self):
        return f"Building(id={self.gdf.id.iloc[0]})"



from dataclasses import dataclass

import pyproj
from lxml import etree

from plateaukit.formats.citygml import CityObject
from plateaukit.formats.citygml.constants import nsmap
from plateaukit.formats.citygml.extractors import city_object_extractors as extractors
from plateaukit.formats.citygml.parsers.geometry_parser import GeometryParser
from plateaukit.formats.citygml.parsers.xml_models.city_object import CityObjectXML

object_type_tags = {
    "Building": f"{{{nsmap['bldg']}}}Building",
    "Road": f"{{{nsmap['tran']}}}Road",
    "Bridge": f"{{{nsmap['brid']}}}Bridge",
}

geometry_type_tags = [
    {"type": "lod0RoofEdge", "tag": f"{{{nsmap['bldg']}}}lod0RoofEdge"},
    {"type": "lod0FootPrint", "tag": f"{{{nsmap['bldg']}}}lod0FootPrint"},
    {"type": "lod1Solid", "tag": f"{{{nsmap['bldg']}}}lod1Solid"},
    {"type": "lod2Solid", "tag": f"{{{nsmap['bldg']}}}lod2Solid"},
    {"type": "lod2MultiSurface", "tag": f"{{{nsmap['bldg']}}}lod2MultiSurface"},
    {"type": "lod1MultiSurface", "tag": f"{{{nsmap['tran']}}}lod1MultiSurface"},
    {"type": "lod2MultiSurface", "tag": f"{{{nsmap['brid']}}}lod2MultiSurface"},
]


@dataclass
class Building(CityObject):
    address: dict | None = None


class CityObjectParser:
    """A parser for CityGML objects.

    Attributes:
        transformer: A pyproj.Transformer instance.
    """

    transformer: pyproj.Transformer | None

    def __init__(
        self,
        transformer: pyproj.Transformer | None = None,
        codelist_map: dict | None = {},
    ):
        self.transformer = transformer
        self.codelist_map = codelist_map or {}


class PLATEAUCityObjectParser(CityObjectParser):
    """A parser for PLATEAU CityGML objects."""

    def _get_geometry(self, root: etree._Element):
        geoms = []

        parser = GeometryParser(transformer=self.transformer)

        for type_tag in geometry_type_tags:
            type = type_tag["type"]
            tag = type_tag["tag"]

            el = root.find(f"./{tag}", nsmap)

            if el is None:
                continue

            if type in ["lod0RoofEdge", "lod0FootPrint"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = {
                    "type": "MultiSurface",
                    "lod": 0,
                    "boundaries": chunked_poslists,
                    "semantics": {
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                }

                geoms.append(geom)
            elif type in ["lod1MultiSurface"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = {
                    "type": "MultiSurface",
                    "lod": 1,
                    "boundaries": chunked_poslists,
                    "semantics": {
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                }

                geoms.append(geom)
            elif type in ["lod1Solid"]:
                # TODO: Fix this
                chunked_poslists = parser.extract_chunked_poslists(el)
                solid_boundaries = [chunked_poslists]

                geom = {
                    "type": "Solid",
                    "lod": 1,
                    "boundaries": solid_boundaries,
                    "semantics": {
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(solid_boundaries))],
                    },
                }

                geoms.append(geom)
            elif type in ["lod2MultiSurface"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = {
                    "type": "MultiSurface",
                    "lod": 2,
                    "boundaries": chunked_poslists,
                    "semantics": {
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                }

                geoms.append(geom)

        # Look through boundedBy
        bound_els = list(root.iterfind("./bldg:boundedBy", nsmap))
        for bound_el in bound_els:
            # TODO: Check semantics
            for type_tag in geometry_type_tags:
                type = type_tag["type"]
                tag = type_tag["tag"]

                el = bound_el.find(f".//{tag}", nsmap)

                if el is None:
                    continue

                if type in ["lod2MultiSurface"]:
                    chunked_poslists = parser.extract_chunked_poslists(el)

                    geom = {
                        "type": "MultiSurface",
                        "lod": 2,
                        "boundaries": chunked_poslists,
                        "semantics": {
                            "surfaces": [
                                {
                                    "type": f"+{type}",
                                }
                            ],
                            "values": [0 for _ in range(len(chunked_poslists))],
                        },
                    }

                    geoms.append(geom)

                else:
                    pass
                    # raise NotImplementedError()

        return geoms

    def _parse_attributes(self, el: CityObjectXML) -> dict:
        # TODO: parse `uro:` attributes

        attributes = dict()

        if el.tree.tag == object_type_tags["Building"]:
            attributes["building_id"] = extractors.get_building_id(el)

            # Optional attributes
            # TODO: Change extractors to transformers
            optional_attributes = {
                "measured_height": extractors.get_measured_height(el),
                "year_of_construction": extractors.get_year_of_construction(el),
                "storeys_above_ground": extractors.get_storeys_above_ground(el),
                "storeys_below_ground": extractors.get_storeys_below_ground(el),
                "name": extractors.get_name(el),
                "usage": extractors.get_usage(el),
            }
            optional_attributes = {
                k: v for k, v in optional_attributes.items() if v is not None
            }
            attributes.update(optional_attributes)

        return attributes

    def parse(self, element: etree._Element) -> CityObject:
        el = CityObjectXML(element, codelist_map=self.codelist_map)

        citygml_id = el.get_gml_id()
        address = None  # self._get_address(el)

        attributes = self._parse_attributes(el)

        if el.tree.tag == object_type_tags["Building"]:
            attributes["building_id"] = extractors.get_building_id(el)

            geometry = self._get_geometry(el.tree)

            obj = Building(
                type="Building",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
                address=address,
            )
        elif el.tree.tag == object_type_tags["Road"]:
            geometry = self._get_geometry(el.tree)

            obj = CityObject(
                type="Road",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
            )
        elif el.tree.tag == object_type_tags["Bridge"]:
            geometry = self._get_geometry(el.tree)

            obj = CityObject(
                type="Bridge",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
            )
        else:
            raise NotImplementedError(f"Unknown object type: {el.tree.tag}")

        return obj



from decimal import Decimal
from itertools import islice
from typing import Any, Iterable


def chunker(it: Iterable, size: int):
    """Turn an iterable into a tuple of chunks in a specified size."""

    iterator = iter(it)
    while chunk := tuple(islice(iterator, size)):
        yield chunk


def parse_posList(text: str):
    """Parse a posList string into a list of 3-chunks."""

    points = text.split(" ")
    points = list(map(lambda x: Decimal(x), points))
    # TODO: fix
    points = list(map(lambda x: float(x), points))
    chunks = list(chunker(points, 3))
    return chunks


def dict_key_to_camel_case(d: dict[str, Any]):
    """Convert dictionary keys to camelCase."""

    def to_camel_case(s):
        t = "".join([w.capitalize() for w in s.split("_")])
        return t[0].lower() + t[1:]

    return {to_camel_case(k): v for k, v in d.items()}



from typing import Any, BinaryIO

from geojson import Feature, FeatureCollection, GeometryCollection, Polygon

from plateaukit.formats.citygml import PLATEAUCityGMLParser
from plateaukit.logger import logger
from plateaukit.utils import dict_key_to_camel_case


def _to_feature(feature_geometry, *, properties: dict[str, Any]):
    _properties = {}

    # _properties["id"] = ...?

    _properties |= dict_key_to_camel_case(properties)

    feat = Feature(
        geometry=feature_geometry,
        properties=_properties,
    )

    return feat


def features_from_gml_single(
    infile: BinaryIO,
    types: list[str] | None = None,
    target_epsg: int = 4326,  # WGS
    altitude: bool = False,
    lod: list[int] = [0],
    codelist_file_map: dict[str, BinaryIO] | None = None,
    attributes: list[str] = ["measuredHeight"],
    allow_geometry_collection: bool = False,
    include_type: bool = False,
):
    """Generate GeoJSON from a single CityGML file."""

    # logger.debug("geojson_from_gml_single")

    parser = PLATEAUCityGMLParser(
        target_epsg=target_epsg, codelist_file_map=codelist_file_map
    )
    co_iter = parser.iterparse(infile)

    # logger.debug(f"citygml: {citygml}")

    if len(lod) > 1:
        raise NotImplementedError("too many LOD values")

    for i, obj in enumerate(co_iter):
        # logger.debug(f"{obj.id}")  # NOTE: Affect performance

        if types and obj.type not in types:
            continue

        geom = None

        if 0 in lod:
            geom = next(filter(lambda x: x["lod"] == 0, obj.geometry), None)
        if 1 in lod:
            geom = next(filter(lambda x: x["lod"] == 1, obj.geometry), None)
            # raise NotImplementedError("LOD 1")
        if 2 in lod:
            geom = next(filter(lambda x: x["lod"] == 2, obj.geometry), None)

        if geom is None:
            continue

        # WIP: exterior only; TODO: Fix this
        base_polygons = [surface[0] for surface in geom["boundaries"]]

        if altitude:
            base_polygons = [
                list(map(lambda x: [x[1], x[0], x[2]], base_polygon))
                for base_polygon in base_polygons
            ]
        else:
            base_polygons = [
                list(map(lambda x: [x[1], x[0]], base_polygon))
                for base_polygon in base_polygons
            ]

        polygons = [Polygon([base_polygon]) for base_polygon in base_polygons]

        properties = dict(obj.attributes) if obj.attributes else {}

        if include_type:
            properties["type"] = obj.type

        if len(polygons) == 1:
            feat = _to_feature(polygons[0], properties=properties)
            yield feat
        else:
            if allow_geometry_collection:
                feat = _to_feature(GeometryCollection(polygons), properties=properties)
                yield feat
            else:
                for polygon in polygons:
                    feat = _to_feature(polygon, properties=properties)
                    yield feat
                # raise AssertionError("No geometry")


def geojson_from_gml_single(
    infile: BinaryIO,
    types: list[str] | None = None,
    target_epsg: int = 4326,  # WGS
    altitude: bool = False,
    lod: list[int] = [0],
    codelist_file_map: dict[str, BinaryIO] | None = None,
    attributes: list[str] = ["measuredHeight"],
    allow_geometry_collection: bool = False,
    include_type: bool = False,
):
    """Generate GeoJSON from a single CityGML file."""

    features = list(
        features_from_gml_single(
            infile,
            types=types,
            target_epsg=target_epsg,
            altitude=altitude,
            lod=lod,
            codelist_file_map=codelist_file_map,
            attributes=attributes,
            allow_geometry_collection=allow_geometry_collection,
            include_type=include_type,
        )
    )
    collection = FeatureCollection(features)

    return collection



from __future__ import annotations

import click

from plateaukit.cli.commands.config import config_cmd
from plateaukit.cli.commands.generate import (
    generate_cityjson_cmd,
    generate_geojson_cmd,
    generate_qmesh_cmd,
)
from plateaukit.cli.commands.info import info_cmd
from plateaukit.cli.commands.install import install_cmd, list_cmd, uninstall_cmd
from plateaukit.cli.commands.prebuild import prebuild_cmd
from plateaukit.logger import set_log_level


# https://alexdelorenzo.dev/notes/click.html
class OrderCommands(click.Group):
    def list_commands(self, ctx):
        return list(self.commands)


@click.group(
    cls=OrderCommands,
    context_settings=dict(help_option_names=["-h", "--help"], show_default=True),
    no_args_is_help=True,
)
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Enable verbose mode."
)
def cli(verbose):
    if verbose:
        set_log_level("DEBUG")


cli.add_command(list_cmd)
cli.add_command(install_cmd)
cli.add_command(uninstall_cmd)
cli.add_command(prebuild_cmd)
cli.add_command(generate_geojson_cmd)
cli.add_command(generate_cityjson_cmd)
cli.add_command(generate_qmesh_cmd)
cli.add_command(info_cmd)
cli.add_command(config_cmd)

if __name__ == "__main__":
    cli()



from __future__ import annotations

from time import sleep
import traceback
from concurrent.futures import Future, ProcessPoolExecutor
from threading import Event

from plateaukit.logger import logger


def wait_futures(
    futures: list[Future],
    pool: ProcessPoolExecutor,
    quit: Event,
    futures_status: dict,
    overall_progress,
    rich_progress,
    shared_progress_status,
):
    def catch_exception(future):
        try:
            future.result()
        except KeyboardInterrupt:
            quit.set()
            pool.shutdown(wait=True, cancel_futures=True)
            raise
        except Exception as e:
            # logger.error(e)
            traceback.print_exc()
            futures_status[future]["failed"] = True

    for future in futures:
        future.add_done_callback(catch_exception)

    try:
        n_futures = len(futures)
        while (n_finished := sum([future.done() for future in futures])) < len(futures):
            for task_id, status in shared_progress_status.items():
                latest = status["progress"]
                total = status["total"]

                if task_id:
                    rich_progress.update(
                        task_id,
                        completed=latest,
                        total=total,
                        visible=latest < total,
                    )

            rich_progress.update(
                overall_progress["task_id"],
                completed=sum(
                    [
                        1
                        for future, f in futures_status.items()
                        if future.done() and f["failed"] is not True
                    ]
                ),
                total=n_futures,
            )

            sleep(0.5)

        for f in futures_status.values():
            if f["task_id"] and f["failed"] is True:
                rich_progress.update(
                    f["task_id"],
                    description=f"[cyan]Progress #{f['counter']} [red]Failed",
                )

        # Finish up the overall progress bar
        rich_progress.update(
            overall_progress["task_id"],
            completed=sum(
                [
                    1
                    for future, f in futures_status.items()
                    if future.done() and f["failed"] is not True
                ]
            ),
            total=n_futures,
            description=f"{overall_progress['description']} [green]Done",
        )

    except KeyboardInterrupt:
        quit.set()
        pool.shutdown(wait=True, cancel_futures=True)
        # pool._processes.clear()
        # concurrent.futures.thread._threads_queues.clear()
        raise



# Based on https://www.geospatial.jp/ckan/dataset/plateau

# %%

city_list = [
    {
        "dataset_id": "plateau-01100-sapporo-shi-2020",
        "city_name": "北海道札幌市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-01100-sapporo-shi-2020",
        "citygml": "5236a116-35be-46ea-a4d0-e840cc7ab3ac",
        "3dtiles": "2306f36e-603e-4be2-8a51-afd4054c7e82",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-01639-sarabetsu-mura-2022",
        "city_name": "北海道更別村",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-01639-sarabetsu-mura-2022",
        "citygml": "35daef49-396e-4c21-9e5c-d602e3a1833b",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-01205-muroran-shi-2022",
        "city_name": "北海道室蘭市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-01205-muroran-shi-2022",
        "citygml": "92bd06f2-7421-4ab9-aeeb-d886887c8a3a",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-02208-mutsu-shi-2022",
        "city_name": "青森県むつ市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-02208-mutsu-shi-2022",
        "citygml": "5f780e16-816d-4239-a345-83c5c0fe8919",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-03201-morioka-shi-2022",
        "city_name": "岩手県盛岡市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-03201-morioka-shi-2022",
        "citygml": "94b62d49-9c3f-4826-8016-8a7816e075cd",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-04100-sendai-shi-2022",
        "city_name": "宮城県仙台市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-04100-sendai-shi-2022",
        "citygml": "a1f35479-9339-4b97-a416-ff120c138db5",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-07204-iwaki-shi-2020",
        "city_name": "福島県いわき市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-07204-iwaki-shi-2020",
        "citygml": "6400e9dc-3dd2-46ad-821f-9072122f44e0",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-07203-koriyama-city-2020",
        "city_name": "福島県郡山市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-07203-koriyama-city-2020",
        "citygml": "ddbc5880-d584-4ce1-bfeb-0f8f46f0135b",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-07205-shirakawa-shi-2020",
        "city_name": "福島県白河市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-07205-shirakawa-shi-2020",
        "citygml": "91c776c5-142a-456c-83d4-7915e8cdb9f6",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-07212-minamisouma-shi-2022",
        "city_name": "福島県南相馬市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-07212-minamisouma-shi-2022",
        "citygml": "2784b256-8649-4375-85d6-ca29ff6de3d6",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-08234-hokota-shi-2020",
        "city_name": "茨城県鉾田市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-08234-hokota-shi-2020",
        "citygml": "02b00898-7370-467e-be92-b7d8a5e00a5a",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-08234-hokota-shi-2022",
        "city_name": "茨城県鉾田市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-08234-hokota-shi-2022",
        "citygml": "1ee20390-7283-47ac-8d74-8db868a136bc",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-08220-tsukuba-shi-2022",
        "city_name": "茨城県つくば市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-08220-tsukuba-shi-2022",
        "citygml": "81551ff2-1680-42a8-a27c-3c01e0bc80ce",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-09201-utsunomiya-shi-2020",
        "city_name": "栃木県宇都宮市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-09201-utsunomiya-shi-2020",
        "citygml": "caa24ca5-5999-4f22-9d8e-8efb9df2cd2c",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-10203-kiryu-shi-2020",
        "city_name": "群馬県桐生市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-10203-kiryu-shi-2020",
        "citygml": "b947726e-c869-4ab9-9483-081ede143038",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-10207-tatebayashi-shi-2020",
        "city_name": "群馬県館林市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-10207-tatebayashi-shi-2020",
        "citygml": "675c8180-e02f-4e5c-ba6c-fc1f47c0ea8f",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-11202-kumagaya-shi-2020",
        "city_name": "埼玉県熊谷市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-11202-kumagaya-shi-2020",
        "citygml": "531c5b69-98b1-4d9f-ad11-89b95c450779",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-11100-saitama-shi-2020",
        "city_name": "埼玉県さいたま市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-11100-saitama-shi-2020",
        "citygml": "114a6d9e-d991-4f19-b686-ebcf5b74719a",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-11100-saitama-shi-2022",
        "city_name": "埼玉県さいたま市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-11100-saitama-shi-2022",
        "citygml": "fbbf84ed-a6a4-4179-8933-f7b144d8e320",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-11224-toda-shi-2022",
        "city_name": "埼玉県戸田市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-11224-toda-shi-2022",
        "citygml": "4550fe74-0eee-4d39-8e35-ddb88898a7a4",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-11230-niiza-shi-2020",
        "city_name": "埼玉県新座市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-11230-niiza-shi-2020",
        "citygml": "d128f7b5-f361-44dd-b28e-e83943a40837",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-11238-hasuda-shi-2022",
        "city_name": "埼玉県蓮田市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-11238-hasuda-shi-2022",
        "citygml": "9397dcb6-1441-4eb7-a803-011d1aea95a5",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-11326-moroyama-machi-2020",
        "city_name": "埼玉県毛呂山町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-11326-moroyama-machi-2020",
        "citygml": "ccdb3ff1-d01b-4669-9d55-6b0ad6da4a33",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-12217-kashiwa-shi-2020",
        "city_name": "千葉県柏市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-12217-kashiwa-shi-2020",
        "citygml": "44e88796-ba94-43a8-b89a-6f41436b97e0",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-12210-mobara-shi-2022",
        "city_name": "千葉県茂原市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-12210-mobara-shi-2022",
        "citygml": "612cd74e-8ed2-41da-ae0c-d176a2322e6a",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-12221-yachiyo-shi-2022",
        "city_name": "千葉県八千代市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-12221-yachiyo-shi-2022",
        "citygml": "f3c84060-a65c-4d25-8d30-f57dbe9aa569",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-tokyo23ku",
        "city_name": "東京都23区",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-tokyo23ku",
        "citygml": "0bab2b7f-6962-41c8-872f-66ad9b40dcb1",
        "3dtiles": "7e2b3b8f-10e0-4f6a-99ec-8ebad021c0d0",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-tokyo23ku-2022",
        "city_name": "東京都23区",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-tokyo23ku-2022",
        "citygml": "55c72dd0-32eb-4107-9526-71fc0af8d50f",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-13229-nishitokyo-shi-2022",
        "city_name": "東京都西東京市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-13229-nishitokyo-shi-2022",
        "citygml": "7b9c2815-e739-4a3d-88cb-66b26106d60b",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-13201-hachioji-shi-2020",
        "city_name": "東京都八王子市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-13201-hachioji-shi-2020",
        "citygml": "4c4fb49e-fcd5-43de-87a1-835aeec0a164",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-13201-hachioji-shi-2022",
        "city_name": "東京都八王子市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-13201-hachioji-shi-2022",
        "citygml": "ef7d1a60-334d-44c6-98c8-e8409779e889",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-13213-higashimurayama-shi-2020",
        "city_name": "東京都東村山市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-13213-higashimurayama-shi-2020",
        "citygml": "4877a936-c244-468f-8bb1-3d3bfa606528",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-14130-kawasaki-shi-2020",
        "city_name": "神奈川県川崎市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-14130-kawasaki-shi-2020",
        "citygml": "51a8a702-03ee-45db-9136-3881daba2477",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-14130-kawasaki-shi-2022",
        "city_name": "神奈川県川崎市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-14130-kawasaki-shi-2022",
        "citygml": "17d4084d-9fed-4194-bc1c-37de9f2a9c7a",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-14150-sagamihara-shi-2020",
        "city_name": "神奈川県相模原市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-14150-sagamihara-shi-2020",
        "citygml": "b910d62c-6a89-4b91-aaec-8a7596432139",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-14382-hakone-machi-2020",
        "city_name": "神奈川県箱根町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-14382-hakone-machi-2020",
        "citygml": "d2a7c149-6519-40a6-a06d-9b6512304b23",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-14201-yokosuka-shi-2020",
        "city_name": "神奈川県横須賀市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-14201-yokosuka-shi-2020",
        "citygml": "31c6a382-f922-42d7-85e0-26ab2c6655ca",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-14100-yokohama-city-2020",
        "city_name": "神奈川県横浜市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-14100-yokohama-city-2020",
        "citygml": "ef70f915-2347-4b50-9008-46a6027bda4b",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-14100-yokohama-shi-2022",
        "city_name": "神奈川県横浜市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-14100-yokohama-shi-2022",
        "citygml": "f4ba72a5-d693-4947-a6ed-547ca346c3c1",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-15100-niigata-shi-2020",
        "city_name": "新潟県新潟市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-15100-niigata-shi-2020",
        "citygml": "1ecc0b56-f9ab-4ab9-b70e-691698a9e4ee",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-15100-niigata-shi-2022",
        "city_name": "新潟県新潟市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-15100-niigata-shi-2022",
        "citygml": "0b0d7ed2-ed57-41f4-b481-11e0454c71a9",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-17206-kaga-shi-2020",
        "city_name": "石川県加賀市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-17206-kaga-shi-2020",
        "citygml": "bf13fbef-e717-446d-a325-25f229e5248d",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-17206-kaga-shi-2021",
        "city_name": "石川県加賀市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-17206-kaga-shi-2021",
        "citygml": "50b61a19-e6bd-437d-9cc1-88eb7042af78",
        "version": "2021",
    },
    {
        "dataset_id": "plateau-17206-kaga-shi-2022",
        "city_name": "石川県加賀市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-17206-kaga-shi-2022",
        "citygml": "dea2a59d-dede-42b9-bc4a-ed270eb3cdc4",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-17201-kanazawa-shi-2020",
        "city_name": "石川県金沢市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-17201-kanazawa-shi-2020",
        "citygml": "87b8b2de-fd92-4710-9f88-d6ccd8693be1",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-20209-ina-shi-2020",
        "city_name": "長野県伊那市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-20209-ina-shi-2020",
        "citygml": "fbaaced1-a87b-405a-aef5-ce1b36e812d8",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-20204-okaya-shi-2020",
        "city_name": "長野県岡谷市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-20204-okaya-shi-2020",
        "citygml": "fbd57aeb-6832-48d1-b5e1-163371d3de2b",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-20204-okaya-shi-2022",
        "city_name": "長野県岡谷市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-20204-okaya-shi-2022",
        "citygml": "59775dd6-0d4b-468a-9581-5c995a5e8bda",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-20217-saku-shi-2022",
        "city_name": "長野県佐久市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-20217-saku-shi-2022",
        "citygml": "6bcf4ae5-86ce-4228-be3c-2f26652102f0",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-20214-chino-shi-2020",
        "city_name": "長野県茅野市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-20214-chino-shi-2020",
        "citygml": "3755ab32-a73d-429d-8583-869850c6053c",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-20214-chino-shi-2022",
        "city_name": "長野県茅野市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-20214-chino-shi-2022",
        "citygml": "45ce6f76-5564-4799-8352-e6e8a7398d03",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-20202-matsumoto-shi-2020",
        "city_name": "長野県松本市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-20202-matsumoto-shi-2020",
        "citygml": "40591402-8363-4512-88ac-3d191a806e8d",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-21201-gifu-shi-2020",
        "city_name": "岐阜県岐阜市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-21201-gifu-shi-2020",
        "citygml": "a8f0da53-862a-4116-a53f-f81690a38d5c",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-21201-gifu-shi-2022",
        "city_name": "岐阜県岐阜市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-21201-gifu-shi-2022",
        "citygml": "08e8f100-7827-4cfa-82e6-f4c8a9665793",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-21211-minokamo-shi-2022",
        "city_name": "岐阜県美濃加茂市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-21211-minokamo-shi-2022",
        "citygml": "3825d2f6-e1de-4e69-bda5-1621aeda2af3",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-19201-kofu-shi-2022",
        "city_name": "山梨県甲府市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-19201-kofu-shi-2022",
        "citygml": "8931fbf1-1873-4406-84da-85f2b1187fd6",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22205-atami-shi-2022",
        "city_name": "静岡県熱海市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22205-atami-shi-2022",
        "citygml": "23c6f893-209f-4ebd-a4be-8a8310f85396",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22222-izu-shi-2022",
        "city_name": "静岡県伊豆市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22222-izu-shi-2022",
        "citygml": "f1b756f2-ccce-4e05-a67e-43d542d7613b",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22225-izunokuni-shi-2022",
        "city_name": "静岡県伊豆の国市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22225-izunokuni-shi-2022",
        "citygml": "db3e4121-9342-4d6e-89ac-4cc376501374",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22208-ito-shi-2022",
        "city_name": "静岡県伊東市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22208-ito-shi-2022",
        "citygml": "6f4a97b6-64b6-474d-b28c-c8fb98d36937",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22211-iwata-shi-2022",
        "city_name": "静岡県磐田市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22211-iwata-shi-2022",
        "citygml": "3f9a656b-36b1-4d0f-b4e2-3a9da72d6c30",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22223-omaezaki-shi-2022",
        "city_name": "静岡県御前崎市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22223-omaezaki-shi-2022",
        "citygml": "0778768e-f2cd-4e67-b298-98a87188f360",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22344-oyama-cho-2022",
        "city_name": "静岡県小山町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22344-oyama-cho-2022",
        "citygml": "f1a32827-15d1-4f3a-b130-23bc3a26c651",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22213-kakegawa-shi-2020",
        "city_name": "静岡県掛川市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22213-kakegawa-shi-2020",
        "citygml": "b6690396-44f5-46e4-90b2-9ba5ab00a388",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22302-kawazu-cho-2022",
        "city_name": "静岡県河津町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22302-kawazu-cho-2022",
        "citygml": "02f2902f-e400-448d-8388-ddc5dcf1768b",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22325-kannami-cho-2022",
        "city_name": "静岡県函南町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22325-kannami-cho-2022",
        "citygml": "53c7d344-c261-42b6-88d2-4ad32efd4c44",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22224-kikugawa-city-2020",
        "city_name": "静岡県菊川市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22224-kikugawa-city-2020",
        "citygml": "f2a8c891-73ff-40f8-a88a-d1415e5a8725",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22221-kosai-shi-2022",
        "city_name": "静岡県湖西市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22221-kosai-shi-2022",
        "citygml": "1c61303d-eebc-4699-9e96-519b38fee92f",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22215-gotenba-shi-2022",
        "city_name": "静岡県御殿場市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22215-gotenba-shi-2022",
        "citygml": "f2187864-6d63-499e-a526-76dcbd2413a7",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22100-shizuoka-shi-2022",
        "city_name": "静岡県静岡市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22100-shizuoka-shi-2022",
        "citygml": "6f0e8388-0dfe-4fcb-aadf-9f41cd26df65",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22341-shimizu-cho-2022",
        "city_name": "静岡県清水町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22341-shimizu-cho-2022",
        "citygml": "f03efa32-d0ce-4689-932f-c88141849bac",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22219-shimoda-shi-2022",
        "city_name": "静岡県下田市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22219-shimoda-shi-2022",
        "citygml": "fda8a9e7-9443-4b0a-87c8-ce963f7eb716",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22220-susono-shi-2022",
        "city_name": "静岡県裾野市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22220-susono-shi-2022",
        "citygml": "7e21dc08-7651-4a5f-95c2-b4ddc30034e0",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22342-nagaizumi-cho-2022",
        "city_name": "静岡県長泉町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22342-nagaizumi-cho-2022",
        "citygml": "7fea2135-bdee-40d2-811e-6c1c7e9eca2a",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22203-numazu-shi-2020",
        "city_name": "静岡県沼津市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22203-numazu-shi-2020",
        "citygml": "71b80915-1c17-4975-a8f2-c69871cee1f5",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-22203-numazu-shi-2021",
        "city_name": "静岡県沼津市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22203-numazu-shi-2021",
        "citygml": "758fe63a-b20f-4d5d-817d-c35eac530143",
        "version": "2021",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22301-higashiizu-cho-2022",
        "city_name": "静岡県東伊豆町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22301-higashiizu-cho-2022",
        "citygml": "407f7b68-0b25-407a-96b8-de176f746e94",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22216-fukuroi-shi-2022",
        "city_name": "静岡県袋井市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22216-fukuroi-shi-2022",
        "citygml": "864d2ca4-045f-490e-ba05-abcbceaeed5d",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22210-fuji-shi-2022",
        "city_name": "静岡県富士市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22210-fuji-shi-2022",
        "citygml": "00da73ef-4aa5-4a04-bd39-ee6bc96a87bd",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22214-fujieda-shi-2022",
        "city_name": "静岡県藤枝市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22214-fujieda-shi-2022",
        "citygml": "0967b2b0-f23b-4d25-85d6-cdc33f2c6b5e",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22207-fujinomiya-shi-2022",
        "city_name": "静岡県富士宮市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22207-fujinomiya-shi-2022",
        "citygml": "75aa2464-afea-44ae-abeb-bc6779895b7f",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22226-makinohara-shi-2022",
        "city_name": "静岡県牧之原市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22226-makinohara-shi-2022",
        "citygml": "3958611b-2ad3-4ce0-a21f-ea78fce3556c",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22206-mishima-shi-2022",
        "city_name": "静岡県三島市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22206-mishima-shi-2022",
        "citygml": "85ec9a31-5960-433e-b352-55928cc63480",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22304-minamiizu-cho-2022",
        "city_name": "静岡県南伊豆町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22304-minamiizu-cho-2022",
        "citygml": "05d34943-457e-46c7-a890-a729e4658916",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22461-mori-machi-2022",
        "city_name": "静岡県森町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22461-mori-machi-2022",
        "citygml": "943ccee9-6b92-45d9-b288-f2f3d3dd31ac",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22212-yaizu-shi-2022",
        "city_name": "静岡県焼津市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22212-yaizu-shi-2022",
        "citygml": "da88169b-3763-4065-9191-f75f6cb03726",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-22424-yoshida-cho-2022",
        "city_name": "静岡県吉田町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-22424-yoshida-cho-2022",
        "citygml": "8d07cf7e-9fdb-4f4c-848c-77c62bcca52d",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-23212-anjo-shi-2020",
        "city_name": "愛知県安城市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-23212-anjo-shi-2020",
        "citygml": "2f3b2351-e7fd-4553-bd00-1424bb05964b",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-23202-okazaki-shi-2020",
        "city_name": "愛知県岡崎市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-23202-okazaki-shi-2020",
        "citygml": "481fde81-ee89-4e1d-a26e-44ec2719d46d",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-23206-kasugai-shi-2022",
        "city_name": "愛知県春日井市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-23206-kasugai-shi-2022",
        "citygml": "199e5bdd-4a2d-479c-8f1a-7d4d75770c10",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-23208-tsushima-shi-2020",
        "city_name": "愛知県津島市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-23208-tsushima-shi-2020",
        "citygml": "18613a1d-16eb-4aa5-aaff-08a40d1aab8b",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-23207-toyokawa-shi-2022",
        "city_name": "愛知県豊川市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-23207-toyokawa-shi-2022",
        "citygml": "809deb42-c22d-498c-afd9-be2249730899",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-23100-nagoya-shi-2020",
        "city_name": "愛知県名古屋市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-23100-nagoya-shi-2020",
        "citygml": "718bf496-7da7-4b7f-83c3-98dd4a8e3acb",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-23230-nisshin-shi-2022",
        "city_name": "愛知県日進市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-23230-nisshin-shi-2022",
        "citygml": "cee5a327-2a72-43ba-935d-4a4bedd86097",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-24212-kumano-shi-2022",
        "city_name": "三重県熊野市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-24212-kumano-shi-2022",
        "citygml": "a63fbbf5-ab09-40d7-9893-31280ee09386",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-24202-yokkaichi-shi-2022",
        "city_name": "三重県四日市市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-24202-yokkaichi-shi-2022",
        "citygml": "f3a2a85c-e1a8-40d8-8c70-c8be6f77ebfa",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-26100-kyoto-shi-2022",
        "city_name": "京都府京都市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-26100-kyoto-shi-2022",
        "citygml": "3da2e055-45a8-4212-bc0e-e769e5f58683",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27204-ikeda-shi-2020",
        "city_name": "大阪府池田市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27204-ikeda-shi-2020",
        "citygml": "8c1b439c-893c-435c-a850-7c91721ede63",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27100-osaka-shi-2020",
        "city_name": "大阪府大阪市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27100-osaka-shi-2020",
        "citygml": "de8fd635-f08c-4143-b22e-c145e695dc43",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-27100-osaka-shi-2022",
        "city_name": "大阪府大阪市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27100-osaka-shi-2022",
        "citygml": "63ba6765-6d33-4f44-9abb-3fa8000c2b3f",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27221-kashiwara-shi-2022",
        "city_name": "大阪府柏原市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27221-kashiwara-shi-2022",
        "citygml": "59662350-46eb-4549-a3b8-29227ad56b85",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27216-kawachinagano-shi-2022",
        "city_name": "大阪府河内長野市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27216-kawachinagano-shi-2022",
        "citygml": "9ae655e2-a0b5-4d8d-8ade-6405abfc2a8e",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27140-sakai-shi-2022",
        "city_name": "大阪府堺市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27140-sakai-shi-2022",
        "citygml": "549d1361-0f67-4ec9-86b8-9577d7684f0d",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27224-settsu-shi-2020",
        "city_name": "大阪府摂津市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27224-settsu-shi-2020",
        "citygml": "a4eaa3a3-8edc-49db-99f9-9815490cc5fe",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27207-takatsuki-shi-2020",
        "city_name": "大阪府高槻市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27207-takatsuki-shi-2020",
        "citygml": "d58a6a37-14dd-4584-a0eb-43c026e705e1",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27341-tadaoka-cho-2020",
        "city_name": "大阪府忠岡町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27341-tadaoka-cho-2020",
        "citygml": "f425d011-2d9c-41e6-bb84-d02e5e4a4b04",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-27203-toyonaka-shi-2020",
        "city_name": "大阪府豊中市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-27203-toyonaka-shi-2020",
        "citygml": "25973afa-4fc8-4d4c-a3a6-f8b2b8efebec",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-28225-asago-shi-2022",
        "city_name": "兵庫県朝来市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-28225-asago-shi-2022",
        "citygml": "c0160066-1c17-4920-a88d-b7bca4443951",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-28210-kakogawa-shi-2020",
        "city_name": "兵庫県加古川市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-28210-kakogawa-shi-2020",
        "citygml": "8ec3a81f-5968-449d-8a65-08a1271bc3f2",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-30422-taiji-cho-2021",
        "city_name": "和歌山県太地町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-30422-taiji-cho-2021",
        "citygml": "021bba37-7720-4eb4-9445-0626ffdbf23b",
        "version": "2021",
        "latest": True,
    },
    {
        "dataset_id": "plateau-30201-wakayama-shi-2022",
        "city_name": "和歌山県和歌山市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-30201-wakayama-shi-2022",
        "citygml": "33c59d3e-0840-4c24-9246-9b11abaedcb0",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-31201-tottori-shi-2020",
        "city_name": "鳥取県鳥取市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-31201-tottori-shi-2020",
        "citygml": "6f8355f0-0d2d-485c-b1c0-85ac9929dd8d",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-34304-kaita-cho-2021",
        "city_name": "広島県海田町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-34304-kaita-cho-2021",
        "citygml": "b7f7ccc1-2008-4cde-8b71-cb673d2a073b",
        "version": "2021",
        "latest": True,
    },
    {
        "dataset_id": "plateau-34202-kure-shi-2020",
        "city_name": "広島県呉市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-34202-kure-shi-2020",
        "citygml": "9f10c73f-c073-40e5-ac14-e429c32c3b15",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-34100-hirosima-shi-2022",
        "city_name": "広島県広島市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-34100-hirosima-shi-2022",
        "citygml": "cd9cf494-84f7-4c59-af3d-ed47d9663dcf",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-34207-fukuyama-shi-2020",
        "city_name": "広島県福山市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-34207-fukuyama-shi-2020",
        "citygml": "b7076aa0-40de-4eab-a863-064ae600ff1d",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-34208-fuchu-shi-2022",
        "city_name": "広島県府中市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-34208-fuchu-shi-2022",
        "citygml": "733b4687-1a4e-415a-84a0-c98f44513814",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-37201-takamatsu-shi-2022",
        "city_name": "香川県高松市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-37201-takamatsu-shi-2022",
        "citygml": "48ac360d-6c98-43d6-aeec-9a7510ca53bb",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-38201-matsuyama-shi-2020",
        "city_name": "愛媛県松山市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-38201-matsuyama-shi-2020",
        "citygml": "6cfb40c0-875a-45ec-a0da-32157af8b15b",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-41423-omachi-cho-2022",
        "city_name": "佐賀県大町町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-41423-omachi-cho-2022",
        "citygml": "2924c864-e79d-407d-9b38-845559733606",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-41208-ogi-shi-2022",
        "city_name": "佐賀県小城市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-41208-ogi-shi-2022",
        "citygml": "18a50707-b4e4-49ce-b1c3-388bbf569fa3",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-41424-kouhoku-machi-2022",
        "city_name": "佐賀県江北町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-41424-kouhoku-machi-2022",
        "citygml": "1d3b5000-305c-4dc4-aebc-277cf225ff30",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-41425-shiroisi-chou-2022",
        "city_name": "佐賀県白石町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-41425-shiroisi-chou-2022",
        "citygml": "f10717be-cd02-4f98-bb72-2c60cb43935c",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-41206-takeo-shi-2022",
        "city_name": "佐賀県武雄市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-41206-takeo-shi-2022",
        "citygml": "71f8ff13-95b7-4dd9-87ee-ffcfcbc32209",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-40205-iizuka-shi-2020",
        "city_name": "福岡県飯塚市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-40205-iizuka-shi-2020",
        "citygml": "34da8e32-7f3b-4a61-87da-2f55d6b563b2",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-40225-ukiha-shi-2022",
        "city_name": "福岡県うきは市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-40225-ukiha-shi-2022",
        "citygml": "ff501855-b467-4d90-9217-48173a469cc4",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-40100-kitakyushu-shi-2020",
        "city_name": "福岡県北九州市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-40100-kitakyushu-shi-2020",
        "citygml": "a393587f-b4bf-4cc9-b5c3-23bbe20572c9",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-40203-kurume-shi-2020",
        "city_name": "福岡県久留米市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-40203-kurume-shi-2020",
        "citygml": "66e8e343-05e8-4850-8cd6-c7c97c432f2e",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-40130-fukuoka-shi-2022",
        "city_name": "福岡県福岡市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-40130-fukuoka-shi-2022",
        "citygml": "f1fa5f41-b876-47bb-b54c-766f0657d1c0",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-40220-munakata-shi-2020",
        "city_name": "福岡県宗像市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-40220-munakata-shi-2020",
        "citygml": "345e8989-2cab-4e26-8174-a344d8ee1b5d",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-43204-arao-shi-2020",
        "city_name": "熊本県荒尾市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-43204-arao-shi-2020",
        "citygml": "f0345e11-2ef0-4600-9b23-402949e43c0f",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-43100-kumamoto-shi-2020",
        "city_name": "熊本県熊本市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-43100-kumamoto-shi-2020",
        "citygml": "211ac99a-b176-4f6a-ad7f-4d7d1d0dcf77",
        "version": "2020",
    },
    {
        "dataset_id": "plateau-43100-kumamoto-shi-2022",
        "city_name": "熊本県熊本市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-43100-kumamoto-shi-2022",
        "citygml": "05b3ba0d-2402-4ca4-99bb-a441323f3b9f",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-43206-tamana-shi-2020",
        "city_name": "熊本県玉名市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-43206-tamana-shi-2020",
        "citygml": "10b6f58a-dbff-4844-963a-c8780fb8bacd",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-43443-mashiki-machi-2020",
        "city_name": "熊本県益城町",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-43443-mashiki-machi-2020",
        "citygml": "a9cfbc79-c00f-4698-95fb-2b8f9a538f60",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-44204-hita-shi-2020",
        "city_name": "大分県日田市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-44204-hita-shi-2020",
        "citygml": "72cac244-9c3e-4ab4-afad-74ad0cc3ad80",
        "version": "2020",
        "latest": True,
    },
    {
        "dataset_id": "plateau-45203-nobeoka-shi-2022",
        "city_name": "宮崎県延岡市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-45203-nobeoka-shi-2022",
        "citygml": "2aca8a01-32e4-406f-a4c4-92f95b99a9c0",
        "version": "2022",
        "latest": True,
    },
    {
        "dataset_id": "plateau-47201-naha-shi-2020",
        "city_name": "沖縄県那覇市",
        "homepage": "https://www.geospatial.jp/ckan/dataset/plateau-47201-naha-shi-2020",
        "citygml": "59819ec2-c617-4508-9827-994cf7b4cae6",
        "version": "2020",
        "latest": True,
    },
]

# import json

# # print(json.dumps(city_list, ensure_ascii=False, indent=2))

# # %%
# import requests

# resp = requests.get("https://www.geospatial.jp/ckan/dataset/plateau")

# # %%
# import re
# from lxml import html

# # print(resp.text)

# root = html.fromstring(resp.text)
# el = root.find_class("embedded-content")[0]
# content = html.tostring(el, encoding="unicode")
# content = content.replace("\n", "")
# # print(content)

# m_all = re.findall(r"strong>(.*?)</strong>(.*?)<p><", content) # TODO: include last one!
# m_all = m_all[3:]

# # print(len(m), m)

# result = []

# for pref_name, pref_part in m_all:
#     # if pref_name == "茨城県":
#     #     print(pref_name, pref_part)
#     m_cities = re.findall(
#         r'<a href="https://www.geospatial.jp/ckan/dataset/(.*?)" title="(.*?)(\d\d\d\d)">.*?</a>',
#         pref_part,
#     )
#     # print(pref_name)
#     # print(m_cities)

#     for dataset_id, name, year in m_cities:
#         result.append(
#             {
#                 "dataset_id": dataset_id,
#                 "city_name": pref_name + name,
#                 "homepage": f"https://www.geospatial.jp/ckan/dataset/{dataset_id}",
#                 # "citygml": None,
#                 # "3dtiles": None,
#             }
#         )

# import json

# print(json.dumps(result, ensure_ascii=False, indent=2))


# # %%

# # diff1 = {x["dataset_id"] for x in city_list} - {x["dataset_id"] for x in result}
# diff2 = {x["dataset_id"] for x in result} - {x["dataset_id"] for x in city_list}
# diff2

# # %%
# from time import sleep

# final_result = []

# for city in result:
#     if city["dataset_id"] not in diff2:
#         city = next(x for x in city_list if x["dataset_id"] == city["dataset_id"])
#         final_result.append(city)
#         continue
#     resp = requests.get(city["homepage"])
#     text = resp.text # .replace("\n", "")

#     ms = re.findall(r'<a class="heading" href=".*?/resource/(.*?)" title="(CityGML.*?)">', text)

#     for i, m in enumerate(ms):
#         print(i, m)

#     if len(ms) == 1:
#         city["citygml"] = ms[take][0]
#         final_result.append(city)
#         sleep(1)
#         continue

#     take = int(input("Which one? "))
#     city["citygml"] = ms[take][0]

#     final_result.append(city)

# print(json.dumps(final_result, ensure_ascii=False, indent=2))
# # for city in city_list:
# #     print({"dataset_id": re.match("https://www.geospatial.jp/ckan/dataset/(.*)", city[1]).group(1), "city_name": city[0], "homepage": city[1]}, ",")

# # for city in city_list:
# #     print(
# #         f'"{city["dataset_id"]}": {{ "city_name": "{city["city_name"]}", "homepage": "{city["homepage"]}"}},'
# #     )

# # %%

# import re

# new_city_list = []

# for city in city_list:
#     try:
#         dataset_id = city["dataset_id"]
#         m = re.match(r".*-(\d\d\d\d)", dataset_id)
#         year = m.group(1)
#         city["version"] = year
#     except:
#         print(city)
#         year = input("year:")
#         city["version"] = year

#     new_city_list.append(city)


# new_city_list

# # %%

# new_city_list = []

# for city in city_list:
#     if [x["city_name"] for x in city_list].count(city["city_name"]) == 1:
#         city["latest"] = True
#     else:
#         cities = [x for x in city_list if x["city_name"] == city["city_name"]]
#         latest_year = max([int(x["version"]) for x in cities])
#         if city["version"] == str(latest_year):
#             city["latest"] = True

#     new_city_list.append(city)


# new_city_list



import json
from collections import defaultdict
from pathlib import Path
from typing import Annotated, Any, DefaultDict

from platformdirs import user_config_dir, user_data_dir
from pydantic import __version__ as pydantic_version

if pydantic_version.startswith("1"):
    from pydantic import BaseModel, Field, PrivateAttr, validator

    class ConfigV1(BaseModel):
        """Class representing the PlateauKit config."""

        path: str = Field(default=None)
        _data_dir: str = PrivateAttr(
            default=str(Path(user_data_dir("plateaukit"), "data"))
        )
        datasets: Any = Field(default_factory=lambda: defaultdict(dict))

        @validator("path", pre=True, always=True)
        @classmethod
        def validate_path(cls, path):
            if path is None:
                config_dir = Path(user_config_dir("plateaukit"))
                if not config_dir.exists():
                    config_dir.mkdir(parents=True)
                path = Path(config_dir, "config.json")

            _path = str(Path(path).resolve())
            return _path

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            # Ensure data directory exists
            Path(self._data_dir).mkdir(parents=True, exist_ok=True)

            # Load config on init
            if Path(self.path).exists():
                with open(self.path, "r") as f:
                    try:
                        config_content = json.load(f)
                    except json.JSONDecodeError:
                        f.seek(0)
                        if f.read() == "":
                            config_content = {}
                        else:
                            raise
            else:
                config_content = {}

            # TODO: fix
            self.datasets = defaultdict(dict, config_content.get("datasets", {}))

        @property
        def data_dir(self):
            return self._data_dir

        def save(self):
            with open(self.path, "w") as f:
                data = self.json(
                    indent=2, exclude_none=True, exclude={"path"}, ensure_ascii=False
                )
                # TODO: remove
                # print("Saved:", data)
                f.write(data)

else:
    from pydantic import BaseModel, Field, PrivateAttr, field_validator

    class ConfigV2(BaseModel):
        """Class representing the PlateauKit config."""

        path: str = Field(default=None, validate_default=True)
        _data_dir: str = PrivateAttr(
            default=str(Path(user_data_dir("plateaukit"), "data"))
        )
        datasets: DefaultDict[str, Annotated[dict, Field(default_factory=dict)]] = (
            Field(default_factory=lambda: defaultdict(dict))
        )

        @field_validator("path", mode="before")
        @classmethod
        def validate_path(cls, path):
            if path is None:
                config_dir = Path(user_config_dir("plateaukit"))
                if not config_dir.exists():
                    config_dir.mkdir(parents=True)
                path = Path(config_dir, "config.json")

            _path = str(Path(path).resolve())
            return _path

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            # Ensure data directory exists
            Path(self._data_dir).mkdir(parents=True, exist_ok=True)

            # Load config on init
            if Path(self.path).exists():
                with open(self.path, "r") as f:
                    try:
                        config_content = json.load(f)
                    except json.JSONDecodeError:
                        f.seek(0)
                        if f.read() == "":
                            config_content = {}
                        else:
                            raise
            else:
                config_content = {}

            # TODO: fix
            self.datasets = defaultdict(dict, config_content.get("datasets", {}))

        @property
        def data_dir(self):
            return self._data_dir

        def save(self):
            with open(self.path, "w") as f:
                data = self.model_dump_json(
                    indent=2, exclude_none=True, exclude={"path"}
                )
                # TODO: remove
                # print("Saved:", data)
                f.write(data)


try:
    Config = ConfigV2
except NameError:
    Config = ConfigV1



import json
from io import StringIO

import geopandas as gpd
import ipydeck
import ipywidgets as widgets
import pandas as pd
import pydeck
from IPython.display import display

import plateaukit.core as core


class InteraciveDeck:
    def __init__(self, gdf: gpd.GeoDataFrame, opacity=1) -> None:
        self.gdf = gdf
        self.deck = self._init_deck(opacity=opacity)

        # Set selection area
        empty_gdf = gpd.GeoDataFrame()
        self.selection = core.area.Area(empty_gdf)

        self._widget_df = self.gdf.copy()

        self.out = widgets.Output()

        def display_df():
            with self.out:
                display(self._widget_df, clear=True)

            return self.out

        self.deck.observe(self.click_handler, names="click")

        self.widget = widgets.VBox([self.deck, self.out])

    def _init_deck(self, opacity=1):
        bbox = self.gdf.total_bounds
        points = [(bbox[0], bbox[1]), (bbox[2], bbox[3])]
        # print(points)

        view_state = pydeck.data_utils.compute_view(points, view_proportion=1)
        # print(view_state, type(view_state))
        view_state.pitch = 45

        view_state = ipydeck.ViewState(
            **{
                "longitude": view_state.longitude,
                "latitude": view_state.latitude,
                "zoom": view_state.zoom,
                "min_zoom": view_state.min_zoom,
                "max_zoom": view_state.max_zoom,
                "pitch": view_state.pitch,
                "bearing": view_state.bearing,
            }
        )

        building_layer = self._create_building_layer(self.gdf, opacity=opacity)

        deck = ipydeck.Deck(
            layers=[building_layer],
            initial_view_state=view_state,
            tooltip={
                "html": "${name ? name : buildingId}",
                "style": {
                    "font-family": "sans-serif",
                    "font-size": "8px",
                    "color": "white",
                },
            },
        )

        return deck

    def _create_building_layer(self, df, opacity=1):
        layer = ipydeck.Layer(
            "GeoJsonLayer",
            data=df,
            filled=True,
            get_fill_color="@@=(properties.fill_color || properties.color || [255, 255, 255])",
            # get_fill_color=[255, 255, 255, 240],
            # get_line_color=[255, 255, 255],
            opacity=opacity,
            extruded=True,
            # wireframe=True,
            get_elevation="@@=properties.measuredHeight",
            pickable=True,
            auto_highlight=True,
            on_click=True,
        )
        return layer

    def click_handler(self, change):
        feature = change["new"]
        # print("clicked", feature)

        if "fill_color" in feature["properties"]:
            del feature["properties"]["fill_color"]

        # NOTE: https://github.com/heremaps/xyz-spaces-python/pull/115
        buf = json.dumps(feature)
        row = gpd.read_file(StringIO(buf))

        if self.selection.gdf.empty:
            self.selection.gdf = row
        else:
            # If self.selection.gdf contains the row, remove it
            if row.at[0, "buildingId"] in self.selection.gdf["buildingId"].values:
                self.selection.gdf = self.selection.gdf[
                    ~self.selection.gdf.loc[:, "buildingId"].isin(row["buildingId"])
                ]
            else:
                self.selection.gdf = pd.concat([self.selection.gdf, row])

        overlap_df = self._widget_df.loc[
            self._widget_df.loc[:, "buildingId"].isin(self.selection.gdf["buildingId"])
        ]

        # Add fill_color column to df if it doesn't exist
        if "fill_color" not in self._widget_df.columns:
            self._widget_df.loc[:, "fill_color"] = None

        self._widget_df["fill_color"] = self._widget_df.loc[
            self._widget_df.index.isin(overlap_df.index), "fill_color"
        ].apply(lambda x: [255, 0, 0])

        updated_layer = self._create_building_layer(self._widget_df)
        self.deck.layers = [updated_layer]

        self.deck.update()

        # Delete fill_color column
        self._widget_df.drop(columns=["fill_color"], inplace=True)

        with self.out:
            # print("clicked", feature)
            display(
                self._widget_df.iloc[self._widget_df.index.isin(overlap_df.index)],
                clear=True,
            )



import json

from plateaukit.logger import logger

from .converter import CityJSONConverter


def cityson_from_gml_serial_with_quit(
    infiles,
    outfile,
    *,
    object_types,
    lod,
    ground,
    codelist_infiles,
    seq=False,
    zipfile=None,
    selection: list[str] | None = None,
    target_epsg: int | None = 3857,
    task_id=None,
    quit=None,
    _progress=None,
    **opts,
):
    # TODO: logger does not work in multiprocessing; must use QueueHandler
    logger.debug("[*] cityson_from_gml_serial_with_quit")

    if not target_epsg:
        target_epsg = 3857  # Web Mercator
        # target_epsg = 4326  # WGS84
        # target_epsg = 32654  # WGS84 / UTM zone 54N
        # target_epsg = 6677  # JGD2011 / Japan Plane Rectangular CS IV

    if seq:
        converter = CityJSONConverter(target_epsg=target_epsg)

        with open(outfile, "w") as f:
            meta = converter.get_meta()
            json.dump(meta, f, ensure_ascii=False, separators=(",", ":"))
            f.write("\n")

            features = converter.features(
                infiles,
                object_types=object_types,
                lod=lod,
                ground=ground,
                codelist_infiles=codelist_infiles,
                zipfile=zipfile,
                selection=selection,
                task_id=task_id,
                quit=quit,
                _progress=_progress,
            )

            for feature in features:
                json.dump(feature, f, ensure_ascii=False, separators=(",", ":"))
                f.write("\n")

    else:
        converter = CityJSONConverter(target_epsg=target_epsg)

        result = converter.convert(
            infiles,
            object_types=object_types,
            lod=lod,
            ground=ground,
            codelist_infiles=codelist_infiles,
            zipfile=zipfile,
            selection=selection,
            task_id=task_id,
            quit=quit,
            _progress=_progress,
        )

        with open(outfile, "w") as f:
            json.dump(result, f, ensure_ascii=False, separators=(",", ":"))

    total = len(infiles) + 1  # + 1 for geojson.dump

    # Complete the progress bar
    if _progress:
        _progress[task_id] = {"progress": total, "total": total}



from __future__ import annotations

import os
import sys
from os import PathLike
from pathlib import Path
from typing import Optional

from plateaukit.config import Config
from plateaukit.download import downloader


def is_dataset_installed(dataset_id, format):
    config = Config()
    path = config.datasets.get(dataset_id, {}).get(format)
    return True if path else False
    # return path and Path(path).exists()


def list_installed_datasets():
    config = Config()
    return list(config.datasets.keys())


def install_dataset(
    dataset_id: str,
    format: str = "citygml",
    local: str | PathLike | None = None,
    force: bool = False,
):
    """Download and install PLATEAU datasets."""

    from plateaukit.download import city_list

    if not dataset_id:
        raise RuntimeError("Missing dataset_id")

    city = next(filter(lambda x: x["dataset_id"] == dataset_id, city_list), None)

    if not city:
        raise RuntimeError("Invalid dataset name")

    installed = is_dataset_installed(dataset_id, format)

    # Abort if a dataset is already installed
    if not force and installed:
        raise RuntimeError(
            f'ERROR: Dataset "{dataset_id}" ({format}) is already installed'
        )

    if local:
        local = Path(local).resolve()

        if not local.exists():
            raise RuntimeError("Local file not found")

        config = Config()
        config.datasets[dataset_id][format] = local
        config.save()

        return
    else:
        # TODO: Check beforehand if specified format is available in city_list
        resource_id = city.get(format, None)

        if not resource_id:
            raise RuntimeError(f"Format '{format}' is not available for '{dataset_id}'")

        config = Config()
        destfile_path = downloader.download_resource(resource_id, dest=config.data_dir)

        config.datasets[dataset_id][format] = destfile_path
        config.save()

        return


def uninstall_dataset(
    dataset_id: str, formats: Optional[list[str]] = None, keep_files: bool = False
):
    """Uninstall PLATEAU datasets."""

    config = Config()
    formats = formats or list(config.datasets[dataset_id].keys())

    if not keep_files:
        paths = []
        for format in formats:
            path = config.datasets[dataset_id].get(format)
            if not path:
                raise RuntimeError(f"Missing files in record for '{format}'")
            paths.append(path)
            try:
                os.remove(path)
            except Exception as e:
                print(e, file=sys.stderr)

    for format in formats:
        if format in config.datasets[dataset_id]:
            del config.datasets[dataset_id][format]

    if len(config.datasets[dataset_id].items()) == 0:
        del config.datasets[dataset_id]

    config.save()



from __future__ import annotations

import click

from plateaukit.prebuild import prebuild


@click.command("prebuild")
@click.argument("dataset_id", nargs=1, required=True)
@click.option(
    "--format",
    type=click.Choice(
        ["parquet", "gpkg"],
        case_sensitive=False,
    ),
    default="parquet",
    help="Internal output format.",
)
@click.option(
    "--split", default=10, help="Split the output into specified number of files."
)
@click.option("-v", "is_verbose", is_flag=True, default=False, help="Verbose")
def prebuild_cmd(dataset_id, format, split, is_verbose):
    """Prebuild PLATEAU datasets.

    PLATEAU データセットを事前ビルドします。
    """

    prebuild(
        dataset_id,
        format=format,
        split=split,
        simple_output=False if is_verbose else True,
    )



from plateaukit.formats.citygml.parsers.city_gml_parser import (
    CityGMLParser,
    PLATEAUCityGMLParser,
)
from plateaukit.formats.citygml.parsers.city_object_parser import (
    CityObjectParser,
    PLATEAUCityObjectParser,
)
from plateaukit.formats.citygml.parsers.codelist_parser import CodelistParser

__all__ = [
    "CityGMLParser",
    "PLATEAUCityGMLParser",
    "CityObjectParser",
    "PLATEAUCityObjectParser",
    "CodelistParser",
]


from __future__ import annotations

import click
from prettytable import PrettyTable

from plateaukit.config import Config
from plateaukit.installer import install_dataset, uninstall_dataset
from plateaukit.prebuild import prebuild


def list_available_datasets(is_all=False):
    from plateaukit.download import city_list

    table = PrettyTable()
    table.field_names = ["id", "name", "version", "homepage"]
    # table.add_row(["all", "(全都市)", "", ""])
    for city in city_list:
        if city.get("latest", False) or is_all:
            table.add_row(
                [
                    city["dataset_id"],
                    city["city_name"],
                    city["version"],
                    city["homepage"],
                ]
            )
    print(table)


@click.command("list")
@click.option(
    "--local", is_flag=True, default=False, help="Show installed datasets only."
)
@click.option(
    "--all",
    is_flag=True,
    default=False,
    help="Show all versions of datasets including old ones.",
)
def list_cmd(local, all):
    """List available and installed PLATEAU datasets.

    利用可能およびインストール済みのデータセットを表示します。
    """
    from plateaukit.download import city_list

    config = Config()

    if local:
        table = PrettyTable()
        table.field_names = ["id", "name", "homepage", "formats"]
        for dataset_id, record in config.datasets.items():
            city = next(
                filter(lambda x: x["dataset_id"] == dataset_id, city_list), None
            )
            if not city:
                continue
            table.add_row(
                [
                    dataset_id,
                    city["city_name"],
                    city["homepage"],
                    " ".join(
                        [x for x in ["citygml", "3dtiles", "gpkg"] if x in record]
                    ),
                ]
            )
        print(table)
        return
    else:
        list_available_datasets(is_all=all)
        return


@click.command("install")
@click.argument("dataset_id", nargs=1, required=False)
@click.option(
    "--format",
    type=click.Choice(["citygml", "3dtiles"], case_sensitive=False),
    default="citygml",
)
@click.option("--local", help="Install local file. (without copying)")
@click.option(
    "--prebuild", "run_prebuild", is_flag=True, default=True, help="Prebuild dataset."
)
@click.option("--force", is_flag=True, default=False, help="Force install.")
@click.option("--download-only", is_flag=True, default=False)
@click.option("-l", "--list", is_flag=True, help="List all latest available datasets.")
@click.option(
    "--list-all", is_flag=True, help="List all available datasets including old ones."
)
@click.option("-v", "is_verbose", is_flag=True, help="Verbose mode.")
def install_cmd(
    dataset_id,
    format,
    local,
    run_prebuild,
    force,
    download_only,
    list,
    list_all,
    is_verbose,
):
    """Download and install PLATEAU datasets.

    PLATEAU データセットをダウンロード・インストールします。
    """
    if not dataset_id and not (list or list_all):
        raise click.UsageError("Missing argument/option: dataset_id or -l/--list")

    if list or list_all:
        list_available_datasets(is_all=list_all)
        return

    if dataset_id:
        try:
            install_dataset(dataset_id, format=format, local=local, force=force)
        except Exception as e:
            raise click.UsageError(str(e))

    if run_prebuild and not download_only:
        prebuild(dataset_id, simple_output=False if is_verbose else True)


@click.command("uninstall")
@click.argument("dataset_id", nargs=1, required=False)
@click.option(
    "--format",
    "-f",
    "formats",
    type=click.Choice(
        ["citygml", "3dtiles", "gpkg"],
        case_sensitive=False,
    ),
    default=None,
    multiple=True,
)
@click.option("--keep-files", is_flag=True, default=False)
def uninstall_cmd(dataset_id, formats: list[str], keep_files):
    """Uninstall PLATEAU datasets.

    PLATEAU データセットをアンインストールします。
    """
    if not dataset_id:
        raise Exception("Missing argument")

    config = Config()
    formats = formats or list(config.datasets[dataset_id].keys())

    # TODO: Fix duplicated code in uninstall_dataset
    if not keep_files:
        paths = []
        for format in formats:
            path = config.datasets[dataset_id].get(format)
            if not path:
                raise RuntimeError(f"Missing files in record for '{format}'")
            paths.append(path)
        click.echo("Would remove:")
        for path in paths:
            click.echo(f"  {path}")
        if click.confirm("Proceed?"):
            uninstall_dataset(dataset_id, formats, keep_files=False)
    else:
        uninstall_dataset(dataset_id, formats, keep_files=True)



import tempfile
from typing import Literal

import ipydeck
import pydeck
from geopandas import GeoDataFrame

from plateaukit.core.widgets.interactive_deck import InteraciveDeck
from plateaukit.logger import logger


class Area:
    """This class represents an area of interest."""

    def __init__(self, gdf: GeoDataFrame) -> None:
        """Initialize an area from a GeoDataFrame.

        Args:
            gdf: GeoDataFrame of the area of interest.
        """
        self.gdf = gdf
        self._datasets: list[str] | None = None

    def __repr__(self) -> str:
        return f"Area()"

    def get_area(self, bbox: list[float] | None = None):
        """Get the specified area from the dataset.

        Args:
            bbox: Bounding box of the area of interest.
                   If not specified, the area of the entire dataset will be returned.
        """

        area_gdf = (
            self.gdf.cx[bbox[0] : bbox[2], bbox[1] : bbox[3]] if bbox else self.gdf
        )

        # TODO: Error handling when area_gdf is empty

        return Area(area_gdf)

    def get_centroid(self) -> list[float]:
        """Get the centroid of the area of interest."""

        bbox = self.gdf.total_bounds
        return [(bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2]

    def pydeck(self, opacity: float = 1):
        # cx, cy = self.get_centroid()

        bbox = self.gdf.total_bounds
        points = [(bbox[0], bbox[1]), (bbox[2], bbox[3])]
        # print(points)

        view_state = pydeck.data_utils.compute_view(points, view_proportion=1)
        view_state.pitch = 45

        deck = pydeck.Deck(
            layers=[
                pydeck.Layer(
                    "GeoJsonLayer",
                    data=self.gdf,
                    filled=True,
                    get_fill_color="fill_color || color || [255, 255, 255]",
                    # get_fill_color=[255, 255, 255, 240],
                    # get_line_color=[255, 255, 255],
                    opacity=opacity,
                    extruded=True,
                    # wireframe=True,
                    get_elevation="measuredHeight",
                    pickable=True,
                    auto_highlight=True,
                )
                # pydeck.Layer(
                #     "PolygonLayer",
                #     data=self.gdf,
                #     get_polygon="geometry.coordinates",
                #     filled=True,
                #     stroked=False,
                #     get_fill_color=[0, 0, 0, 20],
                #     get_line_color=[0, 0, 0, 20],
                # )
            ],
            initial_view_state=view_state,
            # initial_view_state=pydeck.ViewState(
            #     latitude=cy,
            #     longitude=cx,
            #     zoom=1,
            #     pitch=0,
            #     bearing=0,
            # ),
            tooltip={
                "style": {
                    "font-family": "sans-serif",
                    "font-size": "8px",
                    "color": "white",
                },
            },
        )

        return deck

    def ipydeck(self, opacity: float = 1):
        bbox = self.gdf.total_bounds
        points = [(bbox[0], bbox[1]), (bbox[2], bbox[3])]
        # print(points)

        view_state = pydeck.data_utils.compute_view(points, view_proportion=1)
        # print(view_state, type(view_state))
        view_state.pitch = 45

        view_state = ipydeck.ViewState(
            **{
                "longitude": view_state.longitude,
                "latitude": view_state.latitude,
                "zoom": view_state.zoom,
                "min_zoom": view_state.min_zoom,
                "max_zoom": view_state.max_zoom,
                "pitch": view_state.pitch,
                "bearing": view_state.bearing,
            }
        )

        deck = ipydeck.Deck(
            layers=[
                ipydeck.Layer(
                    "GeoJsonLayer",
                    data=self.gdf,
                    filled=True,
                    get_fill_color="@@=(properties.fill_color || properties.color || [255, 255, 255])",
                    # get_fill_color=[255, 255, 255, 240],
                    # get_line_color=[255, 255, 255],
                    opacity=opacity,
                    extruded=True,
                    # wireframe=True,
                    get_elevation="@@=properties.measuredHeight",
                    pickable=True,
                    auto_highlight=True,
                )
            ],
            initial_view_state=view_state,
            # tooltip={
            #     "style": {
            #         "font-family": "sans-serif",
            #         "font-size": "8px",
            #         "color": "white",
            #     },
            # },
        )

        return deck

    def _show_pydeck(self):
        deck = self.pydeck()

        deck._tooltip = {
            "html": "<b>ID:</b> {buildingId}",
            "style": {
                "font-family": "sans-serif",
                "font-size": "8px",
                "color": "white",
            },
        }

        try:
            if __IPYTHON__:  # type: ignore
                return deck.to_html()
        except:
            pass

        with tempfile.NamedTemporaryFile(suffix=".html", delete=False) as fp:
            deck.to_html(fp.name, open_browser=True)

    def _show_ipydeck(self):
        deck = InteraciveDeck(self.gdf)

        self.selection = deck.selection

        return deck.widget

    def show(
        self,
        renderer: Literal["pydeck", "ipydeck"] = "ipydeck",
        # interactive: bool = True,
        embed: bool = True,
    ):
        if renderer == "pydeck":
            return self._show_pydeck()
        elif renderer == "ipydeck":
            return self._show_ipydeck()
        else:
            raise ValueError(f"Unknown renderer: {renderer}")

    def dict(self):
        return self.gdf.to_dict()

    def to_geojson(self):
        """Convert the area to GeoJSON."""

        data = self.gdf.to_json()

        return data

    def to_cityjson(
        self,
        file: str,
        type: list[str] = ["bldg"],
        ground: bool = False,
        target_epsg: int | None = None,
    ):
        """Convert the area to CityJSON."""

        # TODO: Support IOBase as file

        from plateaukit import Dataset

        if self._datasets is None:
            raise RuntimeError("Missing dataset information")

        selection = self.gdf["buildingId"].tolist()

        logger.debug(selection)

        for dataset_id in self._datasets:
            dataset = Dataset(dataset_id)
            dataset.to_cityjson(
                file, ground=ground, selection=selection, target_epsg=target_epsg
            )

    @property
    def buildings(self):
        from plateaukit.core.models import Building

        # TODO: Filter and return buildings only

        for row in self.gdf.itertuples():
            df = self.gdf.loc[[row.Index]]

            yield Building(df)



import numpy as np
from bidict import bidict
from lxml import etree

from plateaukit.constants import nsmap
from plateaukit.logger import logger
from plateaukit.utils import parse_posList


class VerticesMap:
    counter: int
    index_by_vertex: bidict

    def __init__(self):
        self.counter = 0
        self.index_by_vertex = bidict()

    def to_index(self, vertex):
        if vertex not in self.index_by_vertex:
            self.index_by_vertex[vertex] = self.counter
            self.counter += 1
        return self.index_by_vertex[vertex]

    # def to_vertex(self, index):
    #     return self.index_by_vertex.inverse[index]

    # def exists_index(self, index):
    #     return index in self.index_by_vertex.inverse

    @property
    def vertices(self):
        return list(self.index_by_vertex.keys())

    def __contains__(self, vertex):
        return vertex in self.index_by_vertex


def triangles_from_gml(infiles, precision=8):
    try:
        from quantized_mesh_encoder import encode
    except ImportError:
        logger.error("quantized_mesh_encoder is not installed")
        return

    vertices_map = VerticesMap()

    indices = []

    for infile in infiles:
        with open(infile, "r") as f:
            root = etree.parse(f)
            elems_Triangle = root.iterfind(f".//{nsmap['gml']}Triangle")
            for elem_Triangle in elems_Triangle:
                elem_posList = elem_Triangle.find(
                    f"./{nsmap['gml']}exterior/{nsmap['gml']}LinearRing/{nsmap['gml']}posList"
                )
                pos_list = parse_posList(elem_posList.text)
                # Cut off the last element identical to the first
                triangle_vertices = pos_list[:3]
                # print(triangle_vertices)
                triangle_indices = tuple(
                    vertices_map.to_index(v) for v in triangle_vertices
                )
                # print(indices)
                indices.append(triangle_indices)

    vertices = np.array(vertices_map.vertices)
    indices = np.array(indices)
    print(vertices)
    print(indices)

    with open("output.terrain", "wb") as f:
        encode(f, vertices, indices)



from plateaukit.formats.citygml.city_object import CityObject  # noqa: F401
from plateaukit.formats.citygml.citygml import CityGML  # noqa: F401
from plateaukit.formats.citygml.citygml_dataset import CityGMLDataset  # noqa: F401
from plateaukit.formats.citygml.parsers import (
    CityGMLParser,  # noqa: F401
    PLATEAUCityGMLParser,  # noqa: F401
    CityObjectParser,  # noqa: F401
    PLATEAUCityObjectParser,  # noqa: F401
    CodelistParser,  # noqa: F401
)



import io
import os.path
from os import PathLike
from pathlib import Path
from typing import Any, Generator

from bidict import bidict
from fs import open_fs

from plateaukit.formats.citygml import PLATEAUCityGMLParser
from plateaukit.utils import dict_key_to_camel_case

# from json_stream import streamable_dict


class VerticesMap:
    counter: int
    index_by_vertex: bidict

    def __init__(self):
        self.counter = 0
        self.index_by_vertex = bidict()

    def to_index(self, vertex):
        if vertex not in self.index_by_vertex:
            self.index_by_vertex[vertex] = self.counter
            self.counter += 1
        return self.index_by_vertex[vertex]

    # def to_vertex(self, index):
    #     return self.index_by_vertex.inverse[index]

    # def exists_index(self, index):
    #     return index in self.index_by_vertex.inverse

    @property
    def vertices(self):
        return list(self.index_by_vertex.keys())

    def __contains__(self, vertex):
        return vertex in self.index_by_vertex


def _get_nesting_level(l):
    if isinstance(l, list):
        return 1 + max([_get_nesting_level(x) for x in l])
    else:
        return 0


def _shift_to_ground(boundaries: list):
    level = _get_nesting_level(boundaries)

    # Nest level 3
    if level == 3:
        min_z = min(
            [
                min([min([point[2] for point in region]) for region in surface])
                for surface in boundaries
            ]
        )
        boundaries = [
            [
                [(point[0], point[1], point[2] - min_z) for point in region]
                for region in surface
            ]
            for surface in boundaries
        ]
    # Nest level 4
    elif level == 4:
        min_z = min(
            [
                min(
                    [
                        min([min([point[2] for point in shell]) for shell in solid])
                        for solid in surface
                    ]
                )
                for surface in boundaries
            ]
        )
        boundaries = [
            [
                [
                    [(point[0], point[1], point[2] - min_z) for point in shell]
                    for shell in solid
                ]
                for solid in surface
            ]
            for surface in boundaries
        ]

    return boundaries


def get_indexed_boundaries(geometry, vertices_map: VerticesMap, ground=False):
    # TODO: handling composite surface seriously
    # print("get_indexed_boundaries")
    # print("type", geometry["type"])

    boundaries = geometry["boundaries"]
    indexed_boundaries = []

    if geometry["type"] == "MultiPoint":
        for point in geometry["boundaries"]:
            index = vertices_map.to_index(point)
            indexed_boundaries.append(index)
        return indexed_boundaries, vertices_map

    elif geometry["type"] == "MultiLineString":
        for line in geometry["boundaries"]:
            indexed_line = []
            for point in line:
                index = vertices_map.to_index(point)
                indexed_line.append(index)
            indexed_boundaries.append(indexed_line)
        return indexed_boundaries, vertices_map

    elif geometry["type"] in ["MultiSurface", "CompositeSurface"]:
        if ground:
            boundaries = _shift_to_ground(geometry["boundaries"])
        for surface in boundaries:
            # print("surface", surface)
            indexed_surface = []
            for region in surface:
                # print("region", region)

                # NOTE: Reverse the order of vertices; this is necessary for
                # being rendered correctly in cityjson-threejs-loader
                # for SOME datasets
                # eg. plateau-13213-higashimurayama-shi-2020
                ## region = region[::-1]

                indexed_region = []
                unclosed_region = region[:-1]
                for point in unclosed_region:
                    index = vertices_map.to_index(point)
                    indexed_region.append(index)
                indexed_surface.append(indexed_region)
            indexed_boundaries.append(indexed_surface)
        # print("indexed_boundaries", indexed_boundaries, "aaa" * 100)
        return indexed_boundaries, vertices_map

    elif geometry["type"] == "Solid":
        if ground:
            boundaries = _shift_to_ground(geometry["boundaries"])
        for shell in boundaries:
            # print("shell", shell)
            indexed_shell = []
            for surface in shell:
                indexed_surface = []
                for region in surface:
                    # print("region", region)
                    indexed_region = []
                    unclosed_region = region[:-1]
                    for point in unclosed_region:
                        # print("point", point)
                        index = vertices_map.to_index(point)
                        indexed_region.append(index)
                    # print("indexed_region", indexed_region)
                    indexed_surface.append(indexed_region)
                indexed_shell.append(indexed_surface)
            indexed_boundaries.append(indexed_shell)
        # print("indexed_boundaries", indexed_boundaries)
        return indexed_boundaries, vertices_map
    else:
        raise NotImplementedError()

    # TODO: MultiSolid, CompositeSolid

    # for surface in exterior:
    #     unclosed_surface = surface[:-1]
    #     # print("unclosed_surface", unclosed_surface)
    #     single_exterior_surface_exterior = []

    #     for chunk in unclosed_surface:
    #         index = vertices_map.to_index(chunk)
    #         single_exterior_surface_exterior.append(index)

    #     indexed_surface = [single_exterior_surface_exterior]
    #     indexed_surfaces.append(indexed_surface)

    # return indexed_surfaces, vertices_map


class CityJSONConverter:
    def __init__(self, target_epsg: int):
        self.target_epsg = target_epsg

        self.vertices_map = VerticesMap()

    def get_meta(self):
        return {
            "type": "CityJSON",
            "version": "2.0",
            # "extensions": {},
            "transform": {"scale": [1.0, 1.0, 1.0], "translate": [0.0, 0.0, 0.0]},
            "metadata": {
                "referenceSystem": f"https://www.opengis.net/def/crs/EPSG/0/{self.target_epsg}",
            },
            "vertices": [],
            # "appearance": {},
            # "geometry-templates": {},
        }

    def features(
        self,
        infiles: list[str],
        *,
        object_types: list[str] | None,
        lod: list[int],
        ground: bool = False,
        codelist_infiles: list[str] | None = None,
        zipfile: str | PathLike | None = None,
        selection: list[str] | None = None,
        task_id=None,
        quit=None,
        _progress=None,
    ):
        object_iter = self.generate_city_object(
            infiles,
            object_types=object_types,
            lod=lod,
            ground=ground,
            codelist_infiles=codelist_infiles,
            zipfile=zipfile,
            selection=selection,
            task_id=task_id,
            quit=quit,
            _progress=_progress,
        )

        for obj_id, obj in object_iter:
            # vertices = [
            #     transformer.transform(*vertice) for vertice in converter.vertices_map.vertices
            # ]
            vertices = self.vertices_map.vertices

            result = {
                "type": "CityJSONFeature",
                "CityObjects": {obj_id: obj},
                "vertices": vertices,
                # "appearance": {},
                # "geometry-templates": {},
            }

            self.vertices_map = VerticesMap()

            yield result

    def convert(
        self,
        infiles: list[str],
        *,
        object_types: list[str] | None,
        lod: list[int],
        ground: bool = False,
        codelist_infiles: list[str] | None = None,
        zipfile: str | PathLike | None = None,
        selection: list[str] | None = None,
        task_id=None,
        quit=None,
        _progress=None,
    ):
        city_objects = dict(
            self.generate_city_object(
                infiles,
                object_types=object_types,
                lod=lod,
                ground=ground,
                codelist_infiles=codelist_infiles,
                zipfile=zipfile,
                selection=selection,
                task_id=task_id,
                quit=quit,
                _progress=_progress,
            )
        )

        # vertices = [
        #     transformer.transform(*vertice) for vertice in converter.vertices_map.vertices
        # ]
        vertices = self.vertices_map.vertices

        result = {
            "type": "CityJSON",
            "version": "2.0",
            "extensions": {},
            "transform": {"scale": [1.0, 1.0, 1.0], "translate": [0.0, 0.0, 0.0]},
            "metadata": {
                "referenceSystem": f"https://www.opengis.net/def/crs/EPSG/0/{self.target_epsg}",
            },
            "CityObjects": city_objects,
            "vertices": vertices,
            # "appearance": {},
            # "geometry-templates": {},
        }

        return result

    # @streamable_dict
    def generate_city_object(
        self,
        infiles: list[str],
        *,
        object_types: list[str] | None,
        lod: list[int],
        ground: bool = False,
        codelist_infiles: list[str] | None = None,
        zipfile: str | PathLike | None = None,
        selection: list[str] | None = None,
        task_id=None,
        quit=None,
        _progress=None,
    ) -> Generator[tuple[str | None, dict[str, Any]], None, None]:
        # vertices_map = VerticesMap()

        # Load codelists
        codelist_file_map = None

        if zipfile is not None:
            zip_fs = open_fs(f"zip://{zipfile}")
        else:
            zip_fs = None

        if codelist_infiles and infiles:
            base_path = Path(infiles[0]).parent  # TODO: Fix this
            codelist_file_map = dict()

            for codelist_infile in codelist_infiles:
                if zip_fs:
                    with zip_fs.open(codelist_infile, "rb") as f:
                        relative_path = os.path.relpath(codelist_infile, base_path)
                        codelist_file_map[relative_path] = io.BytesIO(f.read())
                else:
                    with open(codelist_infile, "rb") as f:
                        relative_path = os.path.relpath(codelist_infile, base_path)
                        codelist_file_map[relative_path] = io.BytesIO(f.read())

        parser = PLATEAUCityGMLParser(
            target_epsg=self.target_epsg, codelist_file_map=codelist_file_map
        )

        total = len(infiles) + 1  # + 1 for geojson.dump

        for i, infile in enumerate(infiles):
            if task_id is not None and _progress is not None:
                _progress[task_id] = {"progress": i + 1, "total": total}

            _open = zip_fs.open if zip_fs else open

            with _open(infile, "rb") as f:
                co_iter = parser.iterparse(f, selection=selection)
                for city_obj in co_iter:
                    if quit and quit.is_set():
                        return

                    # print("city_obj", city_obj)

                    if object_types is not None and city_obj.type not in object_types:
                        continue

                    indexed_geoms = []

                    if city_obj.geometry is not None:
                        for geom in city_obj.geometry:
                            if geom["lod"] not in lod:
                                continue

                            # print("geom", geom)
                            boundaries, vertices_map = get_indexed_boundaries(
                                geom,
                                self.vertices_map,
                                ground=ground,
                            )

                            indexed_geom = dict(
                                geom, lod=str(geom["lod"]), boundaries=boundaries
                            )

                            # print("indexed_geom", indexed_geom)
                            indexed_geoms.append(indexed_geom)

                            self.vertices_map = vertices_map

                    # print("indexed_geoms", indexed_geoms)

                    # obj_id = city_obj.attributes.get("building_id", city_obj.id)
                    obj_id = city_obj.id

                    yield obj_id, {
                        "type": city_obj.type,  # TODO: There is no Track type etc. in CityJSON
                        "attributes": dict_key_to_camel_case(city_obj.attributes or {}),
                        # "attributes": {"建物ID": "13104-bldg-52530", "measuredHeight": 61.9},
                        # "children": [
                        #     "ID_22730c8f-9fbc-4d58-88dd-5569d7480fad",
                        #     "ID_598f2fab-030f-429c-b938-a222e04d8e4b",
                        #     "ID_db473977-e95e-4075-b0be-55eb65974610",
                        #     "ID_ac26b2cb-553e-428a-9f10-2659419e824d",
                        # ],
                        "geometry": indexed_geoms,
                        # [
                        # {
                        #     "type": "MultiSurface",
                        #     "lod": "0",
                        #     "boundaries": [],
                        # },
                        # {
                        #     "type": "Solid",
                        #     "lod": "1",
                        #     "boundaries": boundaries,
                        #     # "semantics": {
                        #     #     "surfaces": [],
                        #     #     "values": [],
                        #     # },
                        #     # "texture": {"rgbTexture": {"values": []}},
                        # },
                        # ],
                        # "address": [{"Country": "日本", "Locality": "東京都新宿区西新宿一丁目"}],
                    }

        if zip_fs:
            zip_fs.close()
